/*
 * acktable.{cc,hh} --
 *
*/

#include <click/config.h>
#include <click/package.hh>
#include <click/click_ether.h>
#include <click/etheraddress.hh>
#include <click/ipaddress.hh>
#include <click/confparse.hh>
#include <click/bitvector.hh>
#include <click/error.hh>
#include <click/glue.hh>
#include <click/click_adhoc.hh>
#include <click/click_ip.h>
#include <click/click_dsr.hh>
#include <click/router.hh>

#include "acktable.hh"

ACKEntry::ACKEntry()
{
	_ack_id=0;
}

ACKEntry::ACKEntry(int ack_id, struct timeval time_added, Packet *p)
{
	_ack_id=ack_id;
	_time_added.tv_sec=time_added.tv_sec;
	_time_added.tv_usec=time_added.tv_usec;
	_p=p;
}


ACKTable::ACKTable()
         :_expire_timer(expire_hook,this)
{
	MOD_INC_USE_COUNT;
	LIST_INIT(&acache);

	add_input();  /* for IP packet with Source */
	add_input();  /* for DSR ACK packet */
	add_output(); /* for IP packet with source */
	add_output(); /*for IP packet with bad source route */
}

ACKTable::~ACKTable()
{
	MOD_DEC_USE_COUNT;
  	uninitialize();
}

ACKEntry*
ACKTable::findACKEntry(int ack_id)
{
	ACKEntry *a;

	for(a = acache.lh_first; a; a = a->ln_ackentry.le_next)
	{
		if((a->ack_id() ==ack_id))
		{
			return a;
		}
	}
	return 0;
}

void ACKTable::addInfoToACKTable(Packet *p)
{
	const click_dsr_source *dsr_source =(const click_dsr_source *)(p->data()+\
		sizeof(click_ip)+sizeof(click_dsr));
	int hop_count=(dsr_source->dsr_len-2)/4;

	const click_dsr_option * dsr_option =(const click_dsr_option *) (p->data()+\
                  sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_source)+\
                  sizeof(click_dsr_addr)*hop_count);

	const click_dsr_areq *dsr_areq;
	if(dsr_option->dsr_type!=DSR_TYPE_AREQ)
    {
		int offset= sizeof(click_dsr_option)+dsr_option->dsr_len;

		dsr_areq =(const click_dsr_areq *)(p->data()+\
		sizeof(click_ip)+sizeof(click_dsr)+\
		sizeof(click_dsr_source)+\
		hop_count*sizeof(click_dsr_addr)+offset);
    }
    else
    {
		dsr_areq =(const click_dsr_areq *)(p->data()+\
		sizeof(click_ip)+sizeof(click_dsr)+\
		sizeof(click_dsr_source)+\
		hop_count*sizeof(click_dsr_addr));
    }
	struct timeval tv;
	click_gettimeofday(&tv);
	ACKEntry *a;
	int ack_id= ntohs(dsr_areq->dsr_id);

	Packet * p_new=p->clone();
	a=new ACKEntry(ack_id,tv,p_new);
	LIST_INSERT_HEAD(&acache, a, ln_ackentry);
}


void
ACKTable::notify_ninputs(int n)
{
	set_ninputs(n);
}

ACKTable *
ACKTable::clone() const
{
	return new ACKTable;
}

int
ACKTable::configure(const Vector<String> &conf, ErrorHandler *errh)
{
	return 0;
}

int
ACKTable::initialize(ErrorHandler *errh)
{

	/*
		For every element, if it needs to handle timeouts, have a _expire_timer
		element and have a _expire_hook function that is a callback function
		invoked when the timer timeout event occurs
	*/

	_expire_timer.initialize(this);

	/* call the timer event for the first time.. */
	_expire_timer.schedule_after_ms(TIMER_INTERVAL);

	return 0;
}

void
ACKTable::uninitialize()
{
	_expire_timer.unschedule();
}


void
ACKTable::expire_hook(Timer *, void *thunk)
{

	/*
		callback of the timer..this function is called after the timer times out.
		Call all the functions that do the polling checks of the tables here..
	*/

	ACKTable *ack = (ACKTable*)thunk;

	ack->probe_acktable();
}

void
ACKTable::probe_acktable()
{
	ACKEntry *a;
	struct timeval curr_time;
	struct timeval time_added;
	Packet *p;

	click_gettimeofday(&curr_time);

	for(a = acache.lh_first; a; a = a->ln_ackentry.le_next)
	{

		time_added = a->time_added();

		/* Timeout is assumed to be 300 msecs */
		if ( ((curr_time.tv_sec-time_added.tv_sec)*1000000 + \
				(curr_time.tv_usec - time_added.tv_usec ))>=300000 )
		{
			click_chatter("No ACK returned - what are you doing?");
			click_chatter("Remove ACK entry with ID %d\n",a->ack_id());
			p=a->p();
			/*
				Output this packet to routetable input as
				a bad source routed packet
			*/
			output(1).push(p);
			LIST_REMOVE(a,ln_ackentry);
		}
	}
	_expire_timer.schedule_after_ms(TIMER_INTERVAL);
}


void
ACKTable::push(int port, Packet *p_in)
{
	ACKEntry *a;
	int ack_id;

	if ( port == 0)
	{
		/*Check Packet type..if ACK/RRep/Rerr dont add entry..*/

		const click_dsr_option *dsr_option = (const click_dsr_option *)(p_in->data()+\
				sizeof(click_ip)+sizeof(click_dsr));

		if(dsr_option->dsr_type==DSR_TYPE_ACK||dsr_option->dsr_type==DSR_TYPE_RREP||\
					dsr_option->dsr_type==DSR_TYPE_RERR)
			output(0).push(p_in);
		else
		{
			addInfoToACKTable(p_in);
			output(0).push(p_in);
		}
	}
	if ( port == 1 )
	{
		/* Input is an ACK packet. remove the entry for this from the table */
		click_dsr_ack *dsr_ack =(click_dsr_ack *)(p_in->data()+\
					sizeof(click_ip)+sizeof(click_dsr));

		/* Get the value of ack_id */
		ack_id=ntohs(dsr_ack->dsr_id);
		click_chatter("Got an ACK packet with ID %d\n",ack_id);
		a=findACKEntry(ack_id);
		int success=(int)a;
		if(success)
		{
			click_chatter("Removing Entry with ACK Id %d \n",ack_id);
			LIST_REMOVE(a,ln_ackentry);
		}
		else
		{
			click_chatter(" Didnt find the id\n");
		}
		p_in->kill();
	}

}
EXPORT_ELEMENT(ACKTable)




