/*
 * requestable.{cc,hh} --
 *
*/

#include <click/config.h>
#include <click/package.hh>
#include <click/click_ether.h>
#include <click/etheraddress.hh>
#include <click/ipaddress.hh>
#include <click/confparse.hh>
#include <click/bitvector.hh>
#include <click/error.hh>
#include <click/glue.hh>
#include <click/click_adhoc.hh>
#include <click/click_ip.h>
#include <click/click_dsr.hh>
#include <click/router.hh>
#include "requesttable.hh"

RequestEntry::RequestEntry()
{
	_ttl=0;
	_num_rt_disc=0;
	_time_interval=0;
	_request_id=0;
}

RequestEntry::RequestEntry(IPAddress srcIp, IPAddress dstIp, int ttl,\
			    	struct timeval time_last_orig, int num_rt_disc,\
					double time_interval, int request_id)
{
	_srcIp=srcIp;
	_dstIp=dstIp;
	_ttl=ttl;
	_time_last_orig.tv_sec=time_last_orig.tv_sec;
        _time_last_orig.tv_usec=time_last_orig.tv_usec;
	_num_rt_disc=num_rt_disc;
	_time_interval=time_interval;
	_request_id=request_id;
}


RequestTable::RequestTable()
            :_expire_timer(expire_hook,this)
{
	MOD_INC_USE_COUNT;

	_request_id=0;
	me = new IPAddress;

	LIST_INIT(&rcache);
    add_input();  /* For IP packet without Source */
    add_input();  /* For Route Reply */
    add_input();  /* For Route Request Propagation */
	add_output(); /* For DSR Route Request*/
	add_output(); /* For DSR Route Reply*/
}

RequestTable::~RequestTable()
{
	MOD_DEC_USE_COUNT;
  	uninitialize();
}

RequestEntry*
RequestTable::findRequestEntry(IPAddress dst, IPAddress src)
{
	RequestEntry *r;

	for(r = rcache.lh_first; r; r = r->ln_reqentry.le_next)
	{
		if((r->dst_IP() ==dst)&&(r->src_IP()==src))
			return r;
	}
	return 0;
}

RequestEntry*
RequestTable::findRequestEntry(int req_id, IPAddress src)
{
	RequestEntry *r;

	for(r = rcache.lh_first; r; r = r->ln_reqentry.le_next)
	{
		if((r->request_id()==req_id)&&(r->src_IP()==src))
			return r;
	}
	return 0;
}


void RequestTable::addInfoToReqTable(IPAddress srcIp, IPAddress dstIp, int request_id)
{
	struct timeval tv;
	click_gettimeofday(&tv);
	RequestEntry *r;
	r=new RequestEntry(srcIp,dstIp, 1000,tv,1,500, request_id);
	LIST_INSERT_HEAD(&rcache, r, ln_reqentry);
}

Packet *
RequestTable::send_query_for(unsigned src, unsigned dst, int request_id, unsigned char ttl)
{

	/*
		Make a route request packet with room for gratuitious route repair rerrs
	*/

	unsigned payload=sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_rreq)+57;
    WritablePacket *p= Packet::make(payload);

  	/* get header pointers */
	click_ip *ip = reinterpret_cast<click_ip *>(p->data());
	click_dsr *dsr = (click_dsr*)(p->data() + sizeof(click_ip));
	click_dsr_rreq *dsr_rreq = (click_dsr_rreq*)
	    (p->data() + sizeof(click_ip) + sizeof(click_dsr));


	/* Create ip header */
	ip->ip_v = 4;
	ip->ip_hl = sizeof(click_ip) >> 2;
	ip->ip_len = htons(p->length());
	ip->ip_id = htons(request_id);;
	ip->ip_p = IP_PROTO_DSR; // DSR type
	ip->ip_src.s_addr = src;
	ip->ip_dst.s_addr = 0xffffffff;
	ip->ip_tos = 0;
	ip->ip_off = 0;
	ip->ip_ttl = ttl;
	ip->ip_sum = 0;
	ip->ip_sum = click_in_cksum((unsigned char *)ip, sizeof(click_ip));

	/*  create dsr headers */
	/* IMPORTANT: THE DESTINATION HERE IS THE DESTINATION AND NOT
	   BROADCAST
	*/
	dsr->dsr_next_header = 0;
	dsr->dsr_len = htons(sizeof(click_dsr_rreq));
	dsr->dsr_reserved = 0;
	dsr_rreq->dsr_type = DSR_TYPE_RREQ;
	dsr_rreq->dsr_len = 6;
	dsr_rreq->dsr_id = htons(request_id);
	dsr_rreq->dsr_dst.s_addr = dst;



	/* push packet */
	SET_ADHOC_HEADER_ANNO(p, sizeof(click_ip));
	p->set_dst_ip_anno(dst);
	return p;
}

void
RequestTable::notify_ninputs(int n)
{
	set_ninputs(n);
}

Packet * RequestTable::propagate_query(Packet *p_in)
{
	bool is_dsr;
  	const click_dsr *dsr_old=(const click_dsr *)(p_in->data()+\
			sizeof(click_ip));

	const click_dsr_rreq *dsr_rreq_old = (const click_dsr_rreq *)\
				(p_in->data() + \
				sizeof(click_ip) + sizeof(click_dsr));


	int old_hop_count = (dsr_rreq_old->dsr_len - 6) / 4;

	if ( ntohs(dsr_old->dsr_len) > sizeof(click_dsr_rreq) + \
			old_hop_count*sizeof(click_dsr_addr) )
		is_dsr=0; /* Non-DSR */
	else
		is_dsr=1; /* DSR */
	char *buffer;
	int buf_size;
	if (!is_dsr) /* EADSR */
	{
		buf_size = ntohs(dsr_old->dsr_len) - sizeof(click_dsr_rreq)\
				- (old_hop_count * \
				sizeof(click_dsr_addr));

		char *eadsr_opt = (char *)( p_in->data() + sizeof(click_ip) + \
			sizeof(click_dsr) + sizeof(click_dsr_rreq) + \
			(old_hop_count * sizeof(click_dsr_addr)));

		buffer = new char(buf_size);

		memcpy(buffer, eadsr_opt, buf_size);

	}

	WritablePacket *p = static_cast<WritablePacket *>(p_in);

	click_ip *ip = reinterpret_cast<click_ip *>(p->data());

  	click_dsr *dsr=(click_dsr *)(p->data()+\
			sizeof(click_ip));

	click_dsr_rreq *dsr_rreq = (click_dsr_rreq *)(p->data() + \
				    sizeof(click_ip) + sizeof(click_dsr));

	/* Add the address to the end of the packet*/
	p = p_in->put(sizeof(click_dsr_addr));


	/*
		Destination of the route request that has to be propagated
		You have to add yourself to the end and propagate the
   		request
	*/
	IPAddress dst(dsr_rreq->dsr_dst);

	int hop_count = (dsr_rreq->dsr_len - 6) / 4;

	click_dsr_addr *dsr_addr = (click_dsr_addr*)((unsigned char *)dsr_rreq+\
		sizeof(click_dsr_rreq)+hop_count*sizeof(click_dsr_addr));

	dsr_addr->dsr_dst.s_addr = me->addr();
	dsr_rreq->dsr_len += 4;

	if (!is_dsr)
	{
		click_dsr_option *dsr_opt=(click_dsr_option *)((unsigned char *)dsr_rreq+\
				sizeof(click_dsr_rreq)+(hop_count+1)*sizeof(click_dsr_addr));

		memcpy(dsr_opt,buffer,buf_size);
	}

	dsr->dsr_len=htons(ntohs(dsr->dsr_len)+sizeof(click_dsr_addr));
	ip->ip_len = htons(p->length());
	ip->ip_sum = 0;
	ip->ip_sum = click_in_cksum((unsigned char *)ip, sizeof(click_ip));

	p->set_dst_ip_anno(0xffffffff);
	return p;
}

RequestTable *
RequestTable::clone() const
{
	return new RequestTable;
}

int
RequestTable::configure(const Vector<String> &conf, ErrorHandler *errh)
{
        if (cp_va_parse(conf, this, errh, cpKeywords,
                        "IPADDRESS", cpIPAddress, "ip address", me,
                        0)<0)
                return -1;
        return 0;

}

int
RequestTable::initialize(ErrorHandler *errh)
{

	/*
		For every element, if it needs to handle timeouts,
		have a _expire_timer element and have a _expire_hook
		function that is a callback function invoked when the
		timer timeout event occurs
	*/

  _expire_timer.initialize(this);

  _expire_timer.schedule_after_ms(TIMER_HELLO_INTERVAL);

  return 0;
}

void
RequestTable::uninitialize()
{
	_expire_timer.unschedule();
}


void
RequestTable::expire_hook(Timer *, void *thunk)
{
	/*
		Callback of the timer..this function is called after the timer times out.
		Call all the functions that do the polling checks of the tables here..
	*/
  	RequestTable *rt = (RequestTable*)thunk;

	rt->expire_hello();
}

void
RequestTable::expire_hello()
{
	struct timeval curr_time;
	struct timeval time_orig;
	double time_interval;
	click_gettimeofday(&curr_time);

	RequestEntry *r;

	for(r = rcache.lh_first; r; r = r->ln_reqentry.le_next)
	{
		time_orig=r->time_last_orig();

	   /* The lifetime of a rreq entry is 10 secs */
		if((curr_time.tv_sec-time_orig.tv_sec)>=10)
			LIST_REMOVE(r,ln_reqentry);

	}
	/*
		Schedule the next timeout event..can call this here or in the expire_hook function
	*/
	_expire_timer.schedule_after_ms(TIMER_HELLO_INTERVAL);
}


void
RequestTable::push(int port, Packet *p_in)
{
	Packet *p_out;
    const click_ip * ip= p_in->ip_header();

	if(port==0)
	{
		/* Data Packet */
		unsigned dst=(unsigned)p_in->dst_ip_anno();

		unsigned src = ip->ip_src.s_addr;
		IPAddress src_addr(src);
		IPAddress dst_addr(dst);

		/*
			Checks to see if a request for the source and
			destination pair has been sent
		*/
		RequestEntry *r=findRequestEntry(dst_addr,src_addr);
		int success=(int)r;

		p_in->kill();

		/*
			If no request has been sent, increment  request id and
	   		send the request
		*/
		if (success==0)
		{
			_request_id++;
			addInfoToReqTable(src_addr, dst_addr, _request_id);
			p_out=send_query_for(src,dst,_request_id,1000);
			output(0).push(p_out);
		}
		else
		{
			/*
				If no route reply has been received for the previous
				route request, wait for backoff time and re-send
				the route request
			*/
			struct timeval curr_time;
			struct timeval time_orig;
			double time_interval;
			click_gettimeofday(&curr_time);
			time_orig=r->time_last_orig();
			time_interval=r->time_interval();

     		/*
     			If time elapsed >backoff time, send out a RR and double backoff time
     		*/

			if((curr_time.tv_sec-time_orig.tv_sec)*1000000+\
				(curr_time.tv_usec-time_orig.tv_usec)>=time_interval*1000)
			{
				/*Send Next RR*/
				click_chatter(" Need to send next RR for destn\n");
				time_interval=time_interval*2;
				int num_rt_disc=r->num_rt_disc();
				r->set_num_rt_disc(++num_rt_disc);
				r->set_time_interval(time_interval);
				p_out=send_query_for(r->src_IP(),r->dst_IP(),r->request_id(),1000);
				output(0).push(p_out);
			}
			else
				return;
		}

	} /* End of Port 0 processing */

	/*Input is a reply packet..get the id and remove it from the table..*/
	if(port==1)
	{
		const click_dsr_rrep * dsr_rrep=(const click_dsr_rrep *)(p_in->data()+\
							sizeof(click_ip)+sizeof(click_dsr));
		int req_id=ntohs(dsr_rrep->dsr_id);
		IPAddress src(ip->ip_dst.s_addr);

	  	/*find the entry in the request table for that RReq*/
		RequestEntry *r=findRequestEntry(req_id,src);
		int success=(int)r;
		if(success)
			LIST_REMOVE(r,ln_reqentry);

		output(1).push(p_in);
	}

	if(port==2)
	{
		/* If you have to propagate the route request*/
    	unsigned src = ip->ip_src.s_addr;
    	IPAddress src_addr(src);
		const click_dsr_rreq *dsr_rreq = (const click_dsr_rreq*)(p_in->data()\
					 + sizeof(click_ip) + sizeof(click_dsr));

    	IPAddress dst_addr(dsr_rreq->dsr_dst.s_addr);

		int rreq_id= ntohs(dsr_rreq->dsr_id);

		RequestEntry *r=findRequestEntry(rreq_id,src_addr);

		int success = (int)r;

		if (!success)
		{
			/* Add the information to the request table and then
				propagate the route request
			*/
			addInfoToReqTable(src_addr, dst_addr, rreq_id);
			p_out=propagate_query(p_in);
			output(0).push(p_out);
		}
		else
		{
			/* See if we need to propagate the request */
			struct timeval curr_time;
     		struct timeval time_orig;
     		double time_interval;
     		click_gettimeofday(&curr_time);
     		time_orig=r->time_last_orig();
     		time_interval=r->time_interval();

     		/*
     			If time elapsed >backoff time, send out a
		  		RR and double backoff time
		  	*/

     		if((curr_time.tv_sec-time_orig.tv_sec)*1000000+\
		   		(curr_time.tv_usec-time_orig.tv_usec)>=\
				time_interval*1000)
      		{
        		/* Send Next RR */
				time_interval=time_interval*2;
        		int num_rt_disc=r->num_rt_disc();
        		r->set_num_rt_disc(++num_rt_disc);
        		r->set_time_interval(time_interval);
        		p_out=propagate_query(p_in);
        		output(0).push(p_out);
       		}

		}
	} /* If port == 2 */
}
EXPORT_ELEMENT(RequestTable)




