/*
 * routetable.{cc,hh} -- 
 * Carries out the routing table formalities
*/

#include <click/config.h>
#include <click/package.hh>
#include <click/click_ether.h>
#include <click/etheraddress.hh>
#include <click/ipaddress.hh>
#include <click/confparse.hh>
#include <click/bitvector.hh>
#include <click/error.hh>
#include <click/glue.hh>
#include <click/click_dsr.hh>
#include <click/click_adhoc.hh>
#include <click/click_ip.h>
#include <click/router.hh>
#include <click/standard/scheduleinfo.hh>
#include "routetable.hh"
#include <iostream>
#include <fstream>
#include <string>

//Constructor
RouteTable::RouteTable()
          :_expire_timer(expire_hook, this)
{
        //Allocate memory for the current nodes IP Address 
        me=new IPAddress;
        
        //Reset The acknowledgement ID to 0
	_ack_id = 0;

        /* 	Create a LinkCache Object..
        	Assign node id to linkcache..
	*/
	MyLinkCache = new LinkCache(1);

        //Create a send buffer object
        MySendBuffer=new SendBuffer;       

        //Create a Maptable object 
        MyMapTable =new MapTable;

	/* Task to probe the sendbuffer once link cache is changed*/
       t=new Task((RouteTable *)this);

	MOD_INC_USE_COUNT;

       //Inputs

	add_input(); /* for ip packets - Input 0 */
	add_input(); /* for incoming DSR Source routed  packets - Input 1 */
        add_input(); /* For incoming DSR Route Requests - Input 2*/
        add_input(); /* For  Bad DSR Route Packets - Input 3*/

       //Outputs
	add_output(); /* for  source routed ip packets  - Output 0 */
	add_output(); /* for ip packets to the kernel- Output 1 */
        add_output(); /* for no source available IP Packets Output 2*/
        add_output(); /* for propagating DSR route requests  Output 3*/
}

//Standard Click functions for elements

RouteTable::~RouteTable()
{
  MOD_DEC_USE_COUNT;
  uninitialize();
}

void
RouteTable::notify_ninputs(int n)
{
  set_ninputs(n);
}

RouteTable *
RouteTable::clone() const
{
  return new RouteTable;
}

int
RouteTable::configure(const Vector<String> &conf, ErrorHandler *errh)
{
       //Read the parameters from a configuration string 
	if (cp_va_parse(conf, this, errh, cpKeywords,
		        "IPADDRESS", cpIPAddress, "ip address", me, 
			0)<0)
		return -1;
        //Add the current node entry
	MyMapTable->addEntry(*me);
	return 0;
}

int
RouteTable::initialize(ErrorHandler *errh)
{
  
	/*
	  For every element, if it needs to handle 
	  timeouts, have a _expire_timer element and 
	  have a _expire_hook function that is a callback 
	  function invoked when the timer timeout event 
	  occurs
	*/

	ScheduleInfo::initialize_task(this,t,false,errh);
  	_expire_timer.initialize(this);

	/* 
	   This is for probing the sendbuffer to drop packets 
	   whoose lifetime  is over.
	*/
  	_expire_timer.schedule_after_ms(TIMER_INTERVAL);

  	return 0;
}

void
RouteTable::uninitialize()
{
  _expire_timer.unschedule();
  
}


void
RouteTable::expire_hook(Timer *, void *thunk)
{

//callback of the timer..this function is called after the timer times out.
//Call all the functions that do the polling checks of the tables here..

  RouteTable *rt = (RouteTable*)thunk;


 //Check the send buffer for packets that can be sent and drop expired packets
  rt->probe_sendbuffer();
}

void
RouteTable::probe_sendbuffer()
{
  
	SendBufferEntry *s;
  	struct timeval curr_time;
  	struct timeval time_added;
        
        //Get the current time
  	click_gettimeofday(&curr_time);
  
        //Look through all entries in the send buffer		
	for(s=MySendBuffer->sendbuffcache.lh_first;s;\
	    s=s->ln_sendbuffentry.le_next)
	{
	
		/* Get the time difference between the time at which
		   the packet was added in the send buffer and the
 		   the current time when the send buffer is probed
			*/
    		time_added=s->time_added();
  		click_chatter("Probing sendbuffer\n");
  		click_chatter(" Time added is %d, %d \n",time_added.tv_sec, time_added.tv_usec);
  		click_chatter(" Current Time is %d, %d \n",curr_time.tv_sec, curr_time.tv_usec);

		/* If the time is up for that packet, remove the entry from the
	   		send buffer
		*/
    		if((curr_time.tv_sec-time_added.tv_sec)>=30)
    		{  	
			click_chatter("Need to drop this packet\n");
       			LIST_REMOVE(s,ln_sendbuffentry);
    		}
    		else 
			/* Clone the packet and send it to the request table 
	   			to initiate a new route request if possible
			*/
    		{
     			Packet *p=s->p();
     			Packet *p_out=p->clone();
     			output(2).push(p_out);
    		}
  	}

	//Schedule the next timeout event..can call this here or in the expire_hook function
 	_expire_timer.schedule_after_ms(TIMER_INTERVAL);
}


void
RouteTable::push(int port, Packet *p_in)
{   
       Packet *p;
       Packet *p_rrep;           
       Packet *p_data;           
       Packet *p_rerr;           
       int i,j;
       const click_ip *ip = p_in->ip_header();
	if (port==0)  //IP Packet..
	{
	       //Get the dest address of the packet
        	unsigned dst=(unsigned)p_in->dst_ip_anno();
                IPAddress addr(dst);
                click_chatter(" * dsr: destination is %s\n",addr.s().cc());

               //Lookup the Node ID using the Maptable function                
               int dest_id=MyMapTable->ConvertIPtoNode(addr);

		if (dest_id == -1)
			dest_id = MyMapTable->addEntry(addr);
               click_chatter(" * destination in node_id is %d\n",dest_id);
               
             
               Path route;
               
             //Search Route for Destination in the Link cache first
               bool success=MyLinkCache->findRoute(dest_id,route,0);   
            
		/* If route found in the link cache, the route
	   	   is returned as a Path
		*/

              if(success==true)//found the route..
              { 
                 click_chatter("*Yahoo...\n");
                 for(int i=0;i<route.length();i++)
                 	click_chatter(" Node ID: %d \n",route[i].node_id);
              
                //Convert Path obtained to PathIPAddress
                PathIPAddress route_ip;
               
                MyMapTable->ConvertPathtoIP(route, route_ip);
                for(int j=0;j<route_ip.length();j++)
                 click_chatter(" Node IP Address: %s \n",route_ip[j].s().cc());

                //Add  Fixed DSR Header + Source Header to packet..
                p=addDSRHeader(p_in,route_ip);

                //Push the source routed packet out              
                output(0).push(p);
              }
             else
             {
               click_chatter("*Don't have any route to %s\n",addr.s().cc());
               //No route available..buffer this packet and send a copy out to 
               // Request Table
               Packet * p_out=p_in->clone();
               MySendBuffer->AddEntry(p_in);
               output(2).push(p_out);         
             }
 
        }
      

   if (port==1)//Incoming packet is a DSR Source Routed packet..
   {

          // Check for packet type..data/reply/rerr
          const click_dsr_option *dsr_option=(const click_dsr_option *)(\
			p_in->data()+sizeof(click_ip)+sizeof(click_dsr));

          //If Route Reply..
          if(dsr_option->dsr_type==DSR_TYPE_RREP)
          {
             click_chatter("*Route Reply recvd*\n");

            // Extract the reply route..
	     PathIPAddress reply_route=extractReplyRoute(p_in);
		
		/* Printing the entire route for debugging */
      	     for(i=0;i<reply_route.length();i++)
       		   click_chatter(" Node IP Address: %s \n",\
		   reply_route[i].s().cc());

		/* We need to convert the IP Addresses to Node ID's to
		   add to the link cache 
		*/
            Path reply_path;
   
            MyMapTable->ConvertIPtoPath(reply_route,reply_path);


            //Add the Reply path to linkcache
            MyLinkCache->addRoute(reply_path);
            click_chatter("Added route to cache\n");
            
            //Now schedule a task that will send out packets in the sendbuff
           // whose destination has been found using the information
	  // from the route reply
           ErrorHandler *errh;
           t->reschedule();
           
            //Check if this reply is for me or for another node..
           //If for me, then kill this reply packet
	  /* I.e If I am the final destination, kill the packet*/
           unsigned dst= ip->ip_dst.s_addr;
           IPAddress dst_addr(dst);
           if(dst_addr==*me)
              p_in->kill();    
           else
           {
           /* if not for me, then do the forwarding...
	   */
            click_chatter(" Need to forward this to destn %s\n",dst_addr.s().cc());
            p_rrep=forwardPacket(p_in);
            output(0).push(p_rrep);
            }
            return;
           }

         //DSR Route Errors
          if(dsr_option->dsr_type==DSR_TYPE_RERR)
          {
            click_chatter(" Recvd Route Err pkt\n");
 
            //Get a pointer to the route error header
	    const click_dsr_rerr *dsr_rerr = (const click_dsr_rerr *)( p_in->data() + \
sizeof(click_ip) + sizeof(click_dsr));
            const click_dsr_addr *dsr_unreach_addr=(const click_dsr_addr*)( p_in->data() + \
				   sizeof(click_ip) + sizeof(click_dsr) + sizeof(click_dsr_rerr));
            //get the bad hops..
            IPAddress bad_src(dsr_rerr->dsr_src.s_addr);
            IPAddress bad_dst(dsr_unreach_addr->dsr_dst.s_addr);
    	    int link1=MyMapTable->ConvertIPtoNode(bad_src);
		if (link1==-1)
			link1=MyMapTable->addEntry(bad_src);
   	    int link2=MyMapTable->ConvertIPtoNode(bad_dst);
		if (link2==-1)
			link2=MyMapTable->addEntry(bad_dst);
            //Now remove the entries from the linkcache
    	    click_chatter(" Remove %d to %d from linkcache\n", link1, link2);   
    	    MyLinkCache->delLink(link1,link2);
            p_in->kill();
            return;

          }
          else
          {
         //For data packets.. - Get the destination Address

		unsigned dst=(unsigned)p_in->dst_ip_anno();
                IPAddress addr(dst);
                click_chatter(" * dsr: incoming pkt destination is %s\n",addr.s().cc());
          
         int no_hops=0; 
        //Check for ACK request..
	/* Get the DSR Header Pointer */
         click_dsr * dsr=(click_dsr *)(p_in->data()+sizeof(click_ip));
          click_chatter(" OPt type is %d\n",dsr_option->dsr_type);

	/* Get the DSR Source Route Pointer */
         click_dsr_source * dsr_source=(click_dsr_source *)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr));

         click_dsr_areq * dsr_areq;
        //use no of hops in source route to get to the areq header
       //  DSR DRAFT FORMULA!!

         no_hops= (dsr_source->dsr_len-2)/4;
          click_chatter(" No of hops is %d\n",no_hops);

         if(ntohs(dsr->dsr_len)==(sizeof(click_dsr_source)+no_hops*sizeof(click_dsr_addr)+sizeof(click_dsr_areq)))
         { 
	/* Get the pointer to the ACK Request portion of the header*/
         dsr_areq=(click_dsr_areq *)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_source)+no_hops*sizeof(click_dsr_addr));
        }

        else
        {
           //Skip over optional header if there exists one
          const click_dsr_option *dsr_option_new=(const click_dsr_option *)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_source)+sizeof(click_dsr_addr)*no_hops);
          click_chatter(" OPt type is %d\n",dsr_option_new->dsr_type);
       int offset=dsr_option_new->dsr_len+sizeof(click_dsr_option);
        click_chatter("OPt len is %d\n",dsr_option_new->dsr_len);
         dsr_areq=(click_dsr_areq *)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_source)+no_hops*sizeof(click_dsr_addr)+offset);
         }

        //Get the ACK requestor information
        IPAddress ack_src(dsr_areq->dsr_src.s_addr);
        click_chatter(" Received an ACK Request with ID %d for %s\n",ntohs(dsr_areq->dsr_id),ack_src.s().cc()); 
        
        //Create the ACK packet.
        Packet *p_ack=create_dsr_ack(me->addr(),ack_src.addr(),dsr_areq->dsr_id);
       assert(p_ack);
       if(p_ack)
       {
		/* Set the destination address to the source of
		   the node that sent the ACK
		*/
         	p_ack->set_dst_ip_anno(ack_src.addr());
         	output(0).push(p_ack);
       }
      

       //Strip Headers if I'm the destination..
           
       unsigned ip_dst= ip->ip_dst.s_addr;
       IPAddress dst_addr(ip_dst);
       if(dst_addr==*me)
       { 
        	p=StripHeaders(p_in);       
       		//Send it out to the Kernel  
                output(1).push(p);
       }
       else
       {
         click_chatter("Need to forward data packet to destn %s\n",dst_addr.s().cc());

            p_data=forwardDataPacket(p_in);
            output(0).push(p_data);
       }


         }//end for data packets


      } //end for port 1
   
  if(port==2) //Incoming Route Request Packets
    {


      PathIPAddress p_ip;
    //Get the source..
      const click_ip *ip = p_in->ip_header();
      unsigned src= ip->ip_src.s_addr;
      IPAddress src_addr(src);

    //Get the destination of the Route Request
      const click_dsr_rreq * dsr_rreq = (const click_dsr_rreq *)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr));
      IPAddress dst_addr(dsr_rreq->dsr_dst);

	/* ID of the route request */
      int id=ntohs(dsr_rreq->dsr_id);
      click_chatter(" * Got a RReq for destination %s \n",dst_addr.s().cc());
      if(*me==src_addr)
      {
        //I am the source..ignore this rreq
        click_chatter(" I sourced this rreq..ignore..\n");
        p_in->kill();
        return;
      }
      else
	{
                //Save the Route in the RREQ in the linkcache
      		p_ip = createReplyRoute(p_in);
                Path reply_path;
		/* Convert the IP addresses in to Node IDs */
                MyMapTable->ConvertIPtoPath(p_ip,reply_path);
		MyLinkCache->addRoute(reply_path);
	}
      if(*me==dst_addr)
      {  
     		 click_chatter(" * Thats me..have to make RRep now\n");

                 //Reverse the route in the route request
      		 PathIPAddress p_ip_src = p_ip.reverse_route();

      	       	 for(int i=0;i<p_ip.length();i++)
       		     click_chatter(" Node IP Address: %s \n",p_ip[i].s().cc());
                 click_chatter("Rev reply\n");
      	       	 for( int j=0;j<p_ip_src.length();j++)
       		     click_chatter(" Node IP Address: %s \n",p_ip_src[j].s().cc());
                 Path reply_path;
                 MyMapTable->ConvertIPtoPath(p_ip,reply_path);

      	       	 for( int j=0;j<reply_path.length();j++)
       		     click_chatter(" Node ID : %d \n",reply_path[j].node_id);
                 
                 

                 //Create the reply Packet with the dsr header and source
                // header et.al

                 p_rrep=create_dsr_rrep(dst_addr,src_addr,p_ip,p_ip_src,id);
                if(!p_rrep)
                { 
                   click_chatter(" * Didnt create a rep packet..\n");
                 }
                 else
                 {
                   click_chatter(" * Sending out a rep packet..\n");
                   //Push the rrep packet out
                   output(0).push(p_rrep);
                	p_in->kill();
                 }
        } /* I am the destination */
	else
	{
      		//if not me, have to lookup the cache and search for the route..
      		//do a find route here..Do this only if reply from cache is
                // enabled. Currently we dont implement this
      		// if found, then make a cached reply and send it out
     		// if not found, then push the packet out to 
		// the RequestTable element..
                click_chatter("No route for request..passing to req table to propagate\n");
		output(3).push(p_in);
	}
    }

   /* The packet incoming has a bad source route from me to next hop*/
   /* extract this info and remove this link from cache..*/
   if(port==3)
   {
    	IPAddress bad_dst(p_in->dst_ip_anno());
    	IPAddress bad_src(me->addr());
    	click_chatter(" packet incoming has a bad source route \
	from %s to %s ..\n", me->s().cc(), bad_dst.s().cc());

    	int link1=MyMapTable->ConvertIPtoNode(bad_src);
		if (link1==-1)
			link1=MyMapTable->addEntry(bad_src);
    	int link2=MyMapTable->ConvertIPtoNode(bad_dst);
		if (link2==-1)
			link2=MyMapTable->addEntry(bad_dst);
    	click_chatter(" Remove %d to %d from linkcache\n", link1, link2);   

    	MyLinkCache->delLink(link1,link2);
    
    	//get the ip address of the source and check if i'm the source
      	const click_ip *ip = p_in->ip_header();
      	unsigned src= ip->ip_src.s_addr;
      	IPAddress src_addr(src);

	/* If I am the src of the bad source routed packet, kill the packet*/
      	if(src_addr==*me)
    		p_in->kill();
      	else /* Send a route error*/
        {
        	PathIPAddress p_ip;
        	PathIPAddress p_trunc;
        	PathIPAddress source_route;
        	IPAddress my_address;
        	my_address=*me;
        	//Extract source route
        	p_ip=extractSourceRoute(p_in);
        	//truncate till me
        	p_trunc=p_ip.truncate(my_address);
        	//reverse to obtain route from me to destn       
        	source_route=p_trunc.reverse_route();
        	p_rerr=create_dsr_rerr(bad_src,bad_dst,src_addr,source_route);
        	output(0).push(p_rerr);
    		p_in->kill();
       }


   }//end port
 
}
  
Packet * RouteTable::addDSRHeader(Packet *p_in,PathIPAddress source_route)
{
	int old_len=p_in->length();
	int payload=sizeof(click_dsr) + _tailroom;
	int ttl=255;
        int i;
	/* The source and destination addresses are not included
	   as hops in the source route
	*/
        int hop_count=source_route.length()-2;
        WritablePacket *p=static_cast<WritablePacket *>(p_in);

        click_ip *ip= reinterpret_cast<click_ip *>(p->data());
       

	/* We are adding the dsr ack header out here */
       payload += sizeof(click_dsr_source) + \
		  hop_count * sizeof(click_dsr_addr)+ sizeof(click_dsr_areq); 

       click_chatter("*Creating the DSR source routed \
		      packet with the header*\n");

	/* Save the IP Header */
 	click_ip old_ip;
 	memcpy(&old_ip,ip,sizeof(click_ip));   //copy the old header

	//add the extra header size and get a new packet *
 	p=p_in->push(payload);
	
	/* Get the pointer to the head of the data and cast it 
	   as IP Packet Header pointer
	*/
 	ip=reinterpret_cast<click_ip*>(p->data());

  	/* Copy the Old IP Header back     						   into the packet */
	memcpy(ip, & old_ip, sizeof(click_ip)); 

	/* Add the fixed Header */
	click_dsr *dsr = (click_dsr*)(p->data() + sizeof(click_ip));

	dsr->dsr_next_header = ip->ip_p;
	ip->ip_p = IP_PROTO_DSR; /* Protocol type = DSR */
	dsr->dsr_len = htons(payload-sizeof(click_dsr));
	dsr->dsr_reserved = 0;

	click_chatter("*dsr: new packet size is %d, old was %d \n", p->length(), old_len);

	/* Setting IP Header*/
	
	ttl=2; /* ttl=hopcount + 2 - we are right now assuming hopcount to 0 */
	ip->ip_ttl = ttl;
	ip->ip_len = htons(p->length());
	ip->ip_dst.s_addr = (unsigned)p->dst_ip_anno();
        ip->ip_sum=0;
	ip->ip_sum = click_in_cksum((unsigned char *)ip, sizeof(click_ip));

	/* Set Annotations */

	SET_ADHOC_OFFSET_ANNO(p, sizeof(click_ip) + sizeof(click_dsr));
	SET_ADHOC_HEADER_ANNO(p, sizeof(click_ip));

	p->set_ip_header(ip, sizeof(click_ip));

	/* Add Source Header */
        click_dsr_source *dsr_source=(click_dsr_source *)(p->data()+ADHOC_OFFSET_ANNO(p));

       dsr_source->dsr_type=DSR_TYPE_SOURCE_ROUTE;
       dsr_source->dsr_len=4*hop_count+2; 
       dsr_source->dsr_flags=htons(hop_count);

      //add the addresses here now..addr 1..n where full source route is
     // source addr1 ...addr2 ...addr n ..destn..for this case, we dont do 
    //anything as route is only source..destn
      if(hop_count!=0)
       {
         	click_dsr_addr *dsr_source_addr=(click_dsr_addr*)(p->data()+ \
		ADHOC_OFFSET_ANNO(p)+sizeof(click_dsr_source));

		/* Ignore the first and the last address - i.e source and
	   	destination address
		*/
		/* For any given source route, you need to ignore the
		   source, destination and a node before the destination
		   that's why we handle the source length -3
		*/
       		for(i=1;i<=source_route.length()-3 ;i++)
		{
			dsr_source_addr->dsr_dst.s_addr=source_route[i].addr();
			dsr_source_addr++;
		}
		/* for the last hop*/
		dsr_source_addr->dsr_dst.s_addr = source_route[i].addr();
       
       }	
     // set the ip dst anno as the appr next hop here..  
     p->set_dst_ip_anno(source_route[1].addr());      

     SET_ADHOC_OFFSET_ANNO(p, sizeof(click_ip) + \
			   sizeof(click_dsr)+sizeof(click_dsr_source)+\
			   hop_count*sizeof(click_dsr_addr));

     click_chatter("*Added Source Route header*");

	/* Adding the ACK Request Header */

        click_dsr_areq *dsr_areq =(click_dsr_areq *)(p->data()+ADHOC_OFFSET_ANNO(p));

	dsr_areq->dsr_type = DSR_TYPE_AREQ;
	dsr_areq->dsr_len = sizeof(click_dsr_areq) - sizeof(click_dsr_option);
	dsr_areq->dsr_id = htons(_ack_id);
	dsr_areq->dsr_src.s_addr = me->addr() ;

	/* Increment the ack_id by one */
	_ack_id ++;

	SET_ADHOC_OFFSET_ANNO(p, sizeof(click_ip) +  \
		        sizeof(click_dsr)+sizeof(click_dsr_source) + \
		        sizeof(click_dsr_areq));
	return p;
}

Packet * RouteTable::StripHeaders(Packet *p_in)
{
   WritablePacket *p = static_cast<WritablePacket *>(p_in);
   click_ip *ip = reinterpret_cast<click_ip *>(p->data());
   click_dsr *dsr = (click_dsr *)( p->data()+ sizeof(click_ip));

  //Get the length of the DSR headers from the fixed header  

  unsigned dsr_len= sizeof(click_dsr)+ntohs(dsr->dsr_len);
  click_chatter("**Found dsr_len of fixed header as %d \n",ntohs(dsr->dsr_len));
  click_ip new_ip;
  memcpy(&new_ip,ip,sizeof(click_ip));
  new_ip.ip_p=dsr->dsr_next_header;
  //Remove the headers
  
  p->pull(dsr_len);
  memcpy(p->data(),&new_ip,sizeof(click_ip));
  ip=reinterpret_cast<click_ip *>(p->data());
  ip->ip_len=htons(p->length());
  ip->ip_sum=0;
  ip->ip_sum=click_in_cksum((unsigned char *)ip,sizeof(click_ip));

  p->set_ip_header((click_ip*)p->data(),sizeof(click_ip));
 
  click_chatter("* Haha..got out all the headers..\n");
  click_chatter("* Removed %d bytes..\n",dsr_len);
 return p;
}
  

PathIPAddress RouteTable::createReplyRoute(Packet *p_in)
{
  const click_ip *ip = p_in->ip_header();
  unsigned src= ip->ip_src.s_addr;
 
  // Address of the node originating the rreq
  IPAddress src_addr(src);
  const click_dsr_rreq *dsr_rreq=(const click_dsr_rreq *)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr));

//Route Request destination addr
  IPAddress dst_addr(dsr_rreq->dsr_dst);

 int hop_count=(dsr_rreq->dsr_len - 6)/4;
 click_chatter("* Hop count in Rreq is %d\n",hop_count);
// If hop_count is 0, the originator is my immediate neighbor
 if (hop_count==0)
  {
    PathIPAddress p_ip;
    p_ip.appendToPath(src_addr);
    p_ip.appendToPath(dst_addr);
    return p_ip;
  }

 //  case when hop_count>0
 // Route is src..intermediate node 1..node 2...last addr
  if(hop_count>0)
  {
    PathIPAddress p_ip;
    p_ip.appendToPath(src_addr);
    const click_dsr_addr *dsr_addr=(const click_dsr_addr*)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_rreq));
    for(int i=0;i<hop_count-1;i++)
    {
     IPAddress addr(dsr_addr->dsr_dst);
      p_ip.appendToPath(addr);
     dsr_addr++;
    }
    IPAddress last_hop_addr(dsr_addr->dsr_dst);
    p_ip.appendToPath(last_hop_addr);
    p_ip.appendToPath(dst_addr);
    return p_ip;
  }
}

Packet* RouteTable::create_dsr_rrep(IPAddress src,IPAddress dst, 
				    PathIPAddress
				    reply_route, PathIPAddress source_route,					    int id)
{
	WritablePacket *p = NULL;
	/*Creating the payload*/
	int payload = sizeof(click_ip) + sizeof(click_dsr) +\
		      sizeof(click_dsr_rrep)  + \
		      sizeof(click_dsr_addr)* reply_route.length() + \
		      sizeof(click_dsr_source) + \
		      sizeof(click_dsr_addr)*(source_route.length()-2);
       
        //Exclude src and dst in source route for hop count..
	int src_hop_count = source_route.length()-2; 

        //However the reply has to include the source and destn ..
	int reply_hop_count = reply_route.length(); 
	int i;
	unsigned ttl=2;

	/* Make the packet */

	p= Packet::make(payload);
       
        if(!p)
        {
           click_chatter("* Can't make packet  \n");
           return NULL;
        }
	/* getting pointers to the headers */

	click_ip *ip = reinterpret_cast<click_ip *>(p->data());

	click_dsr *dsr = (click_dsr *)(p->data() + sizeof(click_ip));

	click_dsr_rrep *dsr_rrep = (click_dsr_rrep *)( p->data() + \
				   sizeof(click_ip) + sizeof(click_dsr));

	click_dsr_source *dsr_source = (click_dsr_source *)( p->data() + \
				   sizeof(click_ip) + sizeof(click_dsr) + \
				   sizeof(click_dsr_rrep) + \
				   sizeof(click_dsr_addr)* \
				   reply_route.length());

	click_dsr_addr *dsr_source_addr = (click_dsr_addr *)( p->data() + \
				   sizeof(click_ip) + sizeof(click_dsr) + \
				   sizeof(click_dsr_rrep) + \
				   sizeof(click_dsr_addr)* \
				   reply_route.length() + \
				   sizeof(click_dsr_source));
	/* Set Annotations */
	
	SET_ADHOC_OFFSET_ANNO(p,sizeof(click_ip) + sizeof(click_dsr));
	SET_ADHOC_HEADER_ANNO(p,sizeof(click_ip));

	p->set_ip_header(ip, sizeof(click_ip));

	/* Fill the source route header */

	dsr_source->dsr_type = DSR_TYPE_SOURCE_ROUTE;
	dsr_source->dsr_len = 4 * src_hop_count + 2;
	dsr_source->dsr_flags = htons(src_hop_count);

	if (src_hop_count!= 0)
	{
		/*
		  Get a pointer to the addresses and fill them up
		  using source_route addresses
		*/ 
        	for(i=1;i<=source_route.length()-3 ;i++)
		{
			dsr_source_addr->dsr_dst.s_addr=source_route[i].addr();
			dsr_source_addr++;
		}
		/* for the last hop*/
		dsr_source_addr->dsr_dst.s_addr = source_route[i].addr();
	}

	/* Fill the route reply header */
	
	dsr_rrep->dsr_type = DSR_TYPE_RREP;
	dsr_rrep->dsr_len = 4*reply_hop_count+1;
	dsr_rrep->dsr_flags = 0;
	dsr_rrep->dsr_id = htons(id);

	/* Fill the route reply addrs*/

        click_dsr_addr *dsr_addr = (click_dsr_addr*)(p->data()+ \
                                 sizeof(click_ip)+sizeof(click_dsr)+ \
                                 sizeof(click_dsr_rrep));

        for(i=0;i<reply_route.length()-1 ;i++)
	{
		dsr_addr->dsr_dst.s_addr = reply_route[i].addr();
		dsr_addr++;
	}

	/* for the last hop*/
	dsr_addr->dsr_dst.s_addr = reply_route[i].addr();

	/* Make the IP Header */

	ip->ip_v = 4;
	ip->ip_hl = sizeof(click_ip) >> 2;
	ip->ip_len = htons(p->length());
	ip->ip_id = htons(id);
	ip->ip_p = IP_PROTO_DSR;
	ip->ip_src.s_addr = src.addr();
	ip->ip_dst.s_addr = dst.addr();
	ip->ip_tos = 0;
	ip->ip_off = 0;
	ip->ip_ttl = ttl;
	ip->ip_sum = 0;
	ip->ip_sum = click_in_cksum((unsigned char *)ip, sizeof(click_ip));

        
	SET_ADHOC_OFFSET_ANNO(p,payload);

	/* fill the dsr header */
	
	dsr->dsr_next_header = ip->ip_p;
	dsr->dsr_len = htons(payload - sizeof(click_dsr)-sizeof(click_ip));
	dsr->dsr_reserved  = 0;

       //Setting the next hop annotation
       p->set_dst_ip_anno(source_route[1].addr());
       IPAddress dst_anno_address(p->dst_ip_anno());
       click_chatter("*Created RRep Packet with next hop as %s\n",dst_anno_address.s().cc() );
	return p;
}

Packet* RouteTable::create_dsr_rerr(IPAddress bad_src,IPAddress bad_dst,IPAddress src,PathIPAddress source_route )
{
	WritablePacket *p = NULL;
	/*Creating the payload*/
	int payload = sizeof(click_ip) + sizeof(click_dsr) +\
		      sizeof(click_dsr_rerr)  + \
		      sizeof(click_dsr_addr)+ \
		      sizeof(click_dsr_source) + \
		      sizeof(click_dsr_addr)*(source_route.length()-2);
       
        //Exclude src and dst in source route for hop count..
	int src_hop_count = source_route.length()-2; 

       
	int i;
	unsigned ttl=2;

	/* Make the packet */

	p= Packet::make(payload);
       
        if(!p)
        {
           click_chatter("* Can't make packet  \n");
           return NULL;
        }
	/* getting pointers to the headers */

	click_ip *ip = reinterpret_cast<click_ip *>(p->data());

	click_dsr *dsr = (click_dsr *)(p->data() + sizeof(click_ip));

	click_dsr_rerr *dsr_rerr = (click_dsr_rerr *)( p->data() + \
				   sizeof(click_ip) + sizeof(click_dsr));
        click_dsr_addr *dsr_unreach_addr=(click_dsr_addr*)( p->data() + \
				   sizeof(click_ip) + sizeof(click_dsr) + sizeof(click_dsr_rerr));
	click_dsr_source *dsr_source = (click_dsr_source *)( p->data() + \
				   sizeof(click_ip) + sizeof(click_dsr) + \
				   sizeof(click_dsr_rerr) + \
				   sizeof(click_dsr_addr));

	click_dsr_addr *dsr_source_addr = (click_dsr_addr *)( p->data() + \
				   sizeof(click_ip) + sizeof(click_dsr) + \
				   sizeof(click_dsr_rerr) + \
				   sizeof(click_dsr_addr) +\
				   sizeof(click_dsr_source));
	/* Set Annotations */
	
	SET_ADHOC_OFFSET_ANNO(p,sizeof(click_ip) + sizeof(click_dsr));
	SET_ADHOC_HEADER_ANNO(p,sizeof(click_ip));

	p->set_ip_header(ip, sizeof(click_ip));

	/* Fill the source route header */

	dsr_source->dsr_type = DSR_TYPE_SOURCE_ROUTE;
	dsr_source->dsr_len = 4 * src_hop_count + 2;
	dsr_source->dsr_flags = htons(src_hop_count);

	if (src_hop_count!= 0)
	{
		/*
		  Get a pointer to the addresses and fill them up
		  using source_route addresses
		*/ 
        	for(i=1;i<=source_route.length()-3 ;i++)
		{
			dsr_source_addr->dsr_dst.s_addr=source_route[i].addr();
			dsr_source_addr++;
		}
		/* for the last hop*/
		dsr_source_addr->dsr_dst.s_addr = source_route[i].addr();
	}

	/* Fill the route error */
	dsr_rerr->dsr_type = DSR_TYPE_RERR;
        dsr_rerr->dsr_len = 16;
        dsr_rerr->dsr_error = DSR_NODE_UNREACHABLE;
        dsr_rerr->dsr_flags = 0;
        dsr_rerr->dsr_src.s_addr = bad_src.addr();
        dsr_rerr->dsr_dst.s_addr = src.addr();

        // add unreachable destination
	 dsr_unreach_addr->dsr_dst.s_addr = bad_dst.addr();

	/* Make the IP Header */

	ip->ip_v = 4;
	ip->ip_hl = sizeof(click_ip) >> 2;
	ip->ip_len = htons(p->length());
        //id is set to 1..check what value exactly
	ip->ip_id = htons(1);
	ip->ip_p = IP_PROTO_DSR;
	ip->ip_src.s_addr = bad_src.addr();
	ip->ip_dst.s_addr = src.addr();
	ip->ip_tos = 0;
	ip->ip_off = 0;
	ip->ip_ttl = ttl;
	ip->ip_sum = 0;
	ip->ip_sum = click_in_cksum((unsigned char *)ip, sizeof(click_ip));

        
	SET_ADHOC_OFFSET_ANNO(p,payload);

	/* fill the dsr header */
	
	dsr->dsr_next_header = ip->ip_p;
	dsr->dsr_len = htons(payload - sizeof(click_dsr)-sizeof(click_ip));
	dsr->dsr_reserved  = 0;

       //Setting the next hop annotation
       p->set_dst_ip_anno(source_route[1].addr());
       IPAddress dst_anno_address(p->dst_ip_anno());
       click_chatter("*Created Rerr Packet with next hop as %s\n",dst_anno_address.s().cc() );
	return p;
}


PathIPAddress RouteTable::extractReplyRoute(Packet *p_in)
{
 	//Get pointer to the rrep header  
  	const click_dsr_rrep *dsr_rrep=(const click_dsr_rrep *)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr));

	//Get hop count - DRAFT SPECIFIES THE FORMULA!!!
   	int hop_count=(dsr_rrep->dsr_len -1)/4;
   	click_chatter("* Hop count in rrep : %d\n",hop_count);

	//Get pointer to addresses
   	const click_dsr_addr * dsr_addr=(const click_dsr_addr *)(p_in->data()+\
					sizeof(click_ip)+sizeof(click_dsr)+\
					sizeof(click_dsr_rrep));

	//Construct the route from the reply addresses
   	PathIPAddress p_ip;
   	for(int i=0;i<hop_count-1;i++)
   	{
      		IPAddress addr(dsr_addr->dsr_dst);
      		p_ip.appendToPath(addr);
      		dsr_addr++;
   	}
	/* Last address in the reply route */
      	IPAddress addr1(dsr_addr->dsr_dst);
      	p_ip.appendToPath(addr1);

	// Return the route as a PathIPAddress..
   	return p_ip;
}



PathIPAddress RouteTable::extractSourceRoute(Packet *p_in)
{
  //Obtain the pointers to the source route headers
  const click_ip *ip = p_in->ip_header();
  const click_dsr_source *dsr_source=(const click_dsr_source*)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr));
  const click_dsr_addr *dsr_addr=(const click_dsr_addr*)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_source));

 int source_hops= (dsr_source->dsr_len-2)/4;
 PathIPAddress p_ip;

  /* Get the source and the destination from the IP Header */
 IPAddress src(ip->ip_src.s_addr);
 IPAddress dst(ip->ip_dst.s_addr);
 p_ip.appendToPath(src);

	/* Append the intermediate nodes to the path*/
 for(int i=0;i<source_hops-1;i++)
    {
      IPAddress addr(dsr_addr->dsr_dst);
      p_ip.appendToPath(addr);
      dsr_addr++;
    }
	/* last hop! */
  IPAddress addr1(dsr_addr->dsr_dst);
  p_ip.appendToPath(addr1);
  p_ip.appendToPath(dst);
  return p_ip;
}


PathIPAddress RouteTable::reverse_route(PathIPAddress p_ip)
{
 PathIPAddress p_rev;
 for(int i=p_ip.length()-1;i>=0;i--)
 {
  p_rev.appendToPath(p_ip[i]);
 }
  return p_rev;

} 

/* This function is calledd when the join_scheduler function is invoked*/
void RouteTable::run_scheduled()
{
  Packet *p_out;
  SendBufferEntry *s;
  //Look through the send buffer
  for(s=MySendBuffer->sendbuffcache.lh_first;s;s=s->ln_sendbuffentry.le_next)
  {
        Packet *p=s->p();
 	unsigned dst=(unsigned)p->dst_ip_anno();
        IPAddress addr(dst);
        click_chatter(" * dsr: destination is %s\n",addr.s().cc());

       //Lookup the Node ID using the Maptable function                
        int dest_id=MyMapTable->ConvertIPtoNode(addr);
		if (dest_id == -1)
			dest_id = MyMapTable->addEntry(addr);
        click_chatter(" * destination in node_id is %d\n",dest_id);
         //Search Route for Destination in the Link cache first
             
        Path route;
              
        bool success=MyLinkCache->findRoute(dest_id,route,0);   
            
        if(success==true)//found the route..
        { 
        	click_chatter("*Yahoo...\n");
                for(int i=0;i<route.length();i++)
                	click_chatter(" Node ID: %d \n",route[i].node_id);
              
                //Convert Path obtained to PathIPAddress
                PathIPAddress route_ip;
          
                MyMapTable->ConvertPathtoIP(route, route_ip);
                for(int j=0;j<route_ip.length();j++)
                	click_chatter(" Node IP Address: %s \n",\
			route_ip[j].s().cc());

                //Add  Fixed DSR Header + Source Header to packet..
                // To Do :pass the pathIPAddress to this function
                p_out=addDSRHeader(p,route_ip);

                //Push the source routed packet out              
                output(0).push(p_out);
                LIST_REMOVE(s,ln_sendbuffentry);
       }
  }
    
  return;
}


Packet * RouteTable::create_dsr_ack(unsigned src, unsigned dst, unsigned short id)
{
  WritablePacket *q=NULL;

  //make the ACK packet
  q=Packet::make(sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_ack)+64);
 
  //get the pointers to the header.. 
  click_ip *ip=reinterpret_cast<click_ip *>(q->data());
  click_dsr *dsr =(click_dsr*)(q->data()+sizeof(click_ip));
  click_dsr_ack *dsr_ack=(click_dsr_ack*)(q->data()+sizeof(click_ip)+sizeof(click_dsr));


  SET_ADHOC_HEADER_ANNO(q,sizeof(click_ip));
  

  //Fill the dsr headers

  dsr_ack->dsr_type= DSR_TYPE_ACK;
  dsr_ack->dsr_len= sizeof(click_dsr_ack)-sizeof(click_dsr_option);
  dsr_ack->dsr_id=id;
  dsr_ack->dsr_src.s_addr=src;
  dsr_ack->dsr_dst.s_addr=dst;
  dsr->dsr_next_header=0;
  dsr->dsr_reserved =0;
  dsr->dsr_len=htons(dsr_ack->dsr_len + sizeof(click_dsr_option));

  //create the ip header

  ip->ip_v=4;
  ip->ip_hl=sizeof(click_ip)>>2;
  ip->ip_len=htons(q->length());
  ip->ip_id=id;
  ip->ip_p=IP_PROTO_DSR;
  ip->ip_src.s_addr=src;
  ip->ip_dst.s_addr=dst;
  ip->ip_tos=0;
  ip->ip_off=0;
  ip->ip_ttl=255;
  ip->ip_sum=0;
  ip->ip_sum=click_in_cksum((unsigned char *)ip,sizeof(click_ip));
  return q;

}  
  
Packet * RouteTable::forwardPacket(Packet * p_in)
{
 WritablePacket *p=static_cast<WritablePacket*>(p_in);
 click_ip *ip=reinterpret_cast<click_ip *>(p->data());
   

   //Get pointer to the rrep header  
  click_dsr_rrep *dsr_rrep=(click_dsr_rrep *)(p->data()+\
			sizeof(click_ip)+sizeof(click_dsr));

   /*Get hop count*/
   int hop_count=(dsr_rrep->dsr_len -1)/4;
   click_chatter("* Hop count in rrep : %d\n",hop_count);

 	click_dsr_option *dsr_option=(click_dsr_option*)(p->data()+\
				     sizeof(click_ip)+\
				     sizeof(click_dsr)+\
				     sizeof(click_dsr_rrep)+\
				     hop_count*sizeof(click_dsr_addr));
	click_dsr_source *dsr_source;
        //Check if optional header present
	if (dsr_option->dsr_type != DSR_TYPE_SOURCE_ROUTE)
	{
		click_chatter("Not source route \n");
		int offset = sizeof(click_dsr_option) + dsr_option->dsr_len;
		/* Get the source route pointer*/
 		dsr_source = (click_dsr_source*)(p->data()+\
				     sizeof(click_ip)+\
				     sizeof(click_dsr)+\
				     sizeof(click_dsr_rrep)+\
				     hop_count*sizeof(click_dsr_addr)+\
				     offset);
		
	}
	else
	{
		click_chatter("Normal DSR packet \n");
		/* Get the source route pointer*/
 		dsr_source=(click_dsr_source*)(p->data()+\
				     sizeof(click_ip)+\
				     sizeof(click_dsr)+\
				     sizeof(click_dsr_rrep)+\
				     hop_count*sizeof(click_dsr_addr));
	}


	/* This gives you the number of segments left till the 
	   final destination is reached
	   dsr_flags has the number of segments left
	*/
	int segments=ntohs(dsr_source->dsr_flags)-1;

	/* STANDARD FORMULA TO GET THE NUMBER OF HOPS _ FROM DRAFT!!!*/
	int source_hops= (dsr_source->dsr_len-2)/4;
	
	/* The index of the next hop - address to which the packet
           has to be forwarded to
	*/
	int index=source_hops-segments;

	//Update Segments
	dsr_source->dsr_flags=htons(segments);

	/* That means that the next hop is the final destination */
	if(segments==0)
 	{
   		IPAddress final_dst(ip->ip_dst.s_addr);
   		click_chatter("*We are the last hop, have to forward to dest %s\n",final_dst.s().cc());
   		p->set_dst_ip_anno(ip->ip_dst.s_addr);
 	}
	else
 	{
		click_dsr_addr *dsr_source_addr = (click_dsr_addr*)(\
				dsr_source + sizeof(click_dsr_source));
		
   		//get the address of index
   		int i=0;
   		while(i!=index)
   		{
     			dsr_source_addr++;
     			i++;
   		}
   		//set destination annotation to the next hop
   		p->set_dst_ip_anno(dsr_source_addr->dsr_dst.s_addr);
 	}
	return p; 
}

Packet * RouteTable::forwardDataPacket(Packet * p_in)
{
	 WritablePacket *p=static_cast<WritablePacket*>(p_in);
 	 click_ip *ip=reinterpret_cast<click_ip *>(p->data());
 	//Get pointer to the source header  

	 click_dsr_source *dsr_source=(click_dsr_source*)(p->data()+sizeof(click_ip)+sizeof(click_dsr));

	int segments=ntohs(dsr_source->dsr_flags)-1;
	int source_hops= (dsr_source->dsr_len-2)/4;
	int index=source_hops-segments;
	//Update Segments
	dsr_source->dsr_flags=htons(segments);

	if(segments==0)
 	{
   	IPAddress final_dst(ip->ip_dst.s_addr);
   	click_chatter("*We are the last hop, have to forward to dest %s\n",final_dst.s().cc());
   	p->set_dst_ip_anno(ip->ip_dst.s_addr);
 	}
	else
 	{
 	click_dsr_addr *dsr_source_addr=(click_dsr_addr*)(p->data()+sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_source));
   	//get the address of index
   	int i=0;
   	while(i!=index)
   	{
     	dsr_source_addr++;
     	i++;
   	}
   	//set destination annotation to the next hop
   	p->set_dst_ip_anno(dsr_source_addr->dsr_dst.s_addr);
 	}

	//Change the ACK header here..
  
	 click_dsr_option *dsr_opt=(click_dsr_option*)(p->data()+sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_source)+source_hops*sizeof(click_dsr_addr));
 
  	click_dsr_areq *dsr_areq;
  
 	if(dsr_opt->dsr_type!=DSR_TYPE_AREQ)
 	{
    	int offset=sizeof(click_dsr_option)+dsr_opt->dsr_len;    

     	dsr_areq=(click_dsr_areq*)(p->data()+sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_source)+source_hops*sizeof(click_dsr_addr)+offset);
 	}
	else
	{
     	dsr_areq=(click_dsr_areq*)(p->data()+sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_source)+source_hops*sizeof(click_dsr_addr));
 	}
 	dsr_areq->dsr_id = htons(_ack_id);
 	dsr_areq->dsr_src.s_addr = me->addr() ;

  	/* Increment the ack_id by one */
  	_ack_id ++;


	return p;

}


Packet * RouteTable::forwardError(Packet * p_in)
{
 	WritablePacket *p=static_cast<WritablePacket*>(p_in);
 	click_ip *ip=reinterpret_cast<click_ip *>(p->data());
	float send_pwr_dbm;
	int i=0;

	/* Get the source route pointer*/
	click_dsr_source *dsr_source = (click_dsr_source *)( p->data() + \
				   sizeof(click_ip) + sizeof(click_dsr) + \
				   sizeof(click_dsr_rerr) + \
				   sizeof(click_dsr_addr));

	click_dsr_addr *dsr_source_addr = (click_dsr_addr *)( p->data() + \
				   sizeof(click_ip) + sizeof(click_dsr) + \
				   sizeof(click_dsr_rerr) + \
				   sizeof(click_dsr_addr) +\
				   sizeof(click_dsr_source));

	int source_hops= (dsr_source->dsr_len-2)/4;

	

	/* This gives you the number of segments left till the 
	   final destination is reached
	   dsr_flags has the number of segments left
	*/
	int segments=ntohs(dsr_source->dsr_flags)-1;
	click_chatter("The value of segments is %d\n",segments);

	
	/* The index of the next hop - address to which the packet
           has to be forwarded to
	*/
	int index=source_hops-segments;

	//Update Segments
	dsr_source->dsr_flags=htons(segments);

	/* That means that the next hop is the final destination */
	if(segments==0)
 	{
		

   		IPAddress final_dst(ip->ip_dst.s_addr);
   		click_chatter("*We are the last hop, have to forward to \
				dest %s\n",final_dst.s().cc());

   		p->set_dst_ip_anno(ip->ip_dst.s_addr);
 	}
	else
 	{
		/* Get the pointer to the beginning of the addresses*/

   		//get the address of index
   		int i=0;
   		while(i!=index)
   		{
     			dsr_source_addr++;
     	     	
			i++;
   		}
		
   		p->set_dst_ip_anno(dsr_source_addr->dsr_dst.s_addr);
 	}
	return p;
}

bool RouteTable::findMeInRequest(Packet *p_in)
{
         //Get the pointer to the route request header
	  const click_dsr_rreq *dsr_rreq=(const click_dsr_rreq *)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr));

 	int hop_count=(dsr_rreq->dsr_len - 6)/4;
	click_chatter("The value of the hop count in addr reply is %d\n",hop_count);

  	if(hop_count>0)
  	{
    	const click_dsr_addr *dsr_addr=(const click_dsr_addr*)(p_in->data()+sizeof(click_ip)+sizeof(click_dsr)+sizeof(click_dsr_rreq));
    	for(int i=0;i<hop_count;i++)
    		{
     			IPAddress addr(dsr_addr->dsr_dst);
                        //If the current node is in the route request addr list
     			if(*me==addr)
        			return true;       
     			dsr_addr++;
    		}
  	}
   	return false;
}


EXPORT_ELEMENT(RouteTable)
