/*
 * routingtable.{cc,hh} -- the adhoc routing table information element
 */

#include <click/config.h>
#include <click/package.hh>
#include "routingtable.hh"
#include <click/ipaddress.hh>
#include <click/etheraddress.hh>
#include <click/confparse.hh>
#include <click/error.hh>
#include <click/glue.hh>
#include <click/standard/alignmentinfo.hh>
#include <click/router.hh>
#include <click/click_ip.h>
#include <click/sync.hh>
#include <click/llrpc.h>

#ifdef CLICK_SIM
using namespace CLICKSIM;
static String::Initializer crap_initializer; 
#endif

RoutingTable::RoutingTable() 
  : _expire_timer(expire_hook, this), _map(-1)
{
  MOD_INC_USE_COUNT;
}

RoutingTable::~RoutingTable()
{
  MOD_DEC_USE_COUNT;
}

RoutingTable *
RoutingTable::clone() const
{
  return new RoutingTable;
}

int
RoutingTable::initialize(ErrorHandler *errh)
{
  int jiff = click_jiffies();
  int next_timeout_jiffies = -1;

  _seq = 1;
  _id = 0;
  _routes = 0;
  _last_hello_jiffies = jiff;

  _table_lock = new Spinlock();
  if (!_table_lock)
    return errh->error("* rt: cannot create table spinlock");

  _expire_timer.initialize(this);
  // schedule next timeout
  if (_source_lifetime_jiffies == -1)
    next_timeout_jiffies = _lifetime_jiffies;
  else if (_lifetime_jiffies == -1)
    next_timeout_jiffies = _source_lifetime_jiffies;
  else  if (_lifetime_jiffies < _source_lifetime_jiffies)
    next_timeout_jiffies = _lifetime_jiffies;
  else
    next_timeout_jiffies = _source_lifetime_jiffies;
  
  if (next_timeout_jiffies > -1)
    //_expire_timer.schedule_after_ms((1000 * next_timeout_jiffies) / CLICK_HZ);
    _expire_timer.schedule_after_ms(900);

  return 0;
}

void
RoutingTable::uninitialize()
{
  _expire_timer.unschedule();

  delete _table_lock;
}

int
RoutingTable::configure(const Vector<String> &conf, ErrorHandler *errh)
{
  IPAddress ip_src;
  EtherAddress ether_src;
  String addressinfo;
  Vector<String> names;

  // timeout values for route entries
  _lifetime_jiffies = RT_ROUTE_LIFETIME;
  _source_lifetime_jiffies = RT_SOURCE_ROUTE_LIFETIME;

  if (conf.size() < 1 || conf.size() > 3)
    return errh->error("expected NAME1 [NAME2 ... NAMEn] [, SROUTE LIFETIME, ROUTE LIFETIME]");

  if (cp_va_parse(conf, this, errh,
		  cpString, "address info", &addressinfo,
		  cpOptional,
		  cpInteger, "source route lifetime", &_source_lifetime_jiffies,
		  cpOptional,
		  cpInteger, "table driven lifetime", &_lifetime_jiffies,
		  cpEnd) < 0)
    return -1;

  // create list of interfaces from names supplied
  cp_spacevec(addressinfo, names);
  _ip_iflist.resize(names.size());
  _ether_iflist.resize(names.size());

  for (int i = 0; i < names.size(); i++) {
    // copy address to src (will handle aliases)
    cp_ip_address(names[i], &ip_src, this);
    _ip_iflist[i] = ip_src;
    cp_ethernet_address(names[i], &ether_src, this);
    _ether_iflist[i] = ether_src;
    click_chatter("* rt: interface %d is %s\n", i, 
		  ether_src.s().cc());
  }
  
  // convert route lifetimes to jiffies
  if (_lifetime_jiffies > -1)
    _lifetime_jiffies = (CLICK_HZ * _lifetime_jiffies) / 1000;
  if (_source_lifetime_jiffies > -1)
    _source_lifetime_jiffies = (CLICK_HZ * _source_lifetime_jiffies) / 1000;

  return 0;
}  

void
RoutingTable::expire_hook(Timer *timer, void *thunk)
{
  RoutingTable *table = (RoutingTable*)thunk;
  int next_timeout_jiffies = -1;

  // debug
  static int count = 0;

  Vector<unsigned> dst_list;
  // expire route table entries
  table->expire_routes(dst_list);
  // additional protocol callbacks
  table->do_expire_callbacks(timer, dst_list);

  // schedule next timeout
  if (table->_source_lifetime_jiffies == -1)
    next_timeout_jiffies = table->_lifetime_jiffies;
  else if (table->_lifetime_jiffies == -1)
    next_timeout_jiffies = table->_source_lifetime_jiffies;
  else  if (table->_lifetime_jiffies < table->_source_lifetime_jiffies)
    next_timeout_jiffies = table->_lifetime_jiffies;
  else
    next_timeout_jiffies = table->_source_lifetime_jiffies;
  
  if (next_timeout_jiffies > -1)
    //table->_expire_timer.schedule_after_ms((1000 * next_timeout_jiffies) / CLICK_HZ);

  if (count == 800)
    table->packet_count();

  count++;
}

void
RoutingTable::register_expire_callback(ExpireHook hook, Vector<unsigned> *dst_list, void *thunk)
{
  ExpireEntry *eentry = new ExpireEntry;
 
  eentry->hook = hook;
  eentry->dst_list = dst_list;
  eentry->thunk = thunk;

  // todo: should have some way of unregistering...
  _expire_callbacks.push_back(eentry);
}

void
RoutingTable::do_expire_callbacks(Timer *timer, Vector<unsigned> &dst_list)
{
  for (int i = 0; i < _expire_callbacks.size(); i++) {
    ExpireEntry *eentry = _expire_callbacks[i];
    if (eentry) {
      // need to copy list of destinations
      if (eentry->dst_list) {
	eentry->dst_list->clear();
	for (int i = 0; i < dst_list.size(); i++)
	  eentry->dst_list->push_back(dst_list[i]);
      }
      eentry->hook(timer, eentry->thunk);
    }
  }
}

void
RoutingTable::clear_map()
{
  _table_lock->acquire();

  // clear all cache entries
  for (int i = 0; i < _routes; i++) {
    if (RouteInfo *info = _as[i]) {
      // delete queued packet entries
      for (PacketEntry *packet = info->plist_head; packet; ) {
	PacketEntry *next = packet->next;
	if (packet->p)
	  packet->p->kill();
	delete packet;
	packet = next;
      }
      // delete route entries
      for (RouteEntry *route = info->rlist_head; route; ) {
	RouteEntry *next = route->next;
	delete route;
	route = next;
      }
      delete info;
    }
  }

  _as.clear();
  _map.clear();
  _table_lock->release();
}

void
RoutingTable::expire_routes(Vector<unsigned> &dst_list)
{
  int jiff = click_jiffies();
  bool inc_seq = false;
  bool route_break = false;
  int shortest_gap = 1000;

  _table_lock->acquire();

  for (int i = 0; i < _routes; i++) {
    RouteInfo *info = _as[i];
    if (info) {
      IPAddress addr(info->dst);
      IPAddress me(_ip_iflist[0]);
      info->lock->acquire();
      RouteEntry *route = info->rlist_head;
      RouteEntry *prev_route = NULL;
      RouteEntry *next_route = NULL;
      RouteEntry *prev_shortest_route = info->shortest_route;
      while (route) {
	next_route = route->next;
	int gap = jiff - route->last_response_jiffies;
	// debug
	IPAddress addr(info->dst);
	IPAddress me(_ip_iflist[0].s().cc());
	IPAddress match("192.168.1.34");
	IPAddress match2("192.168.1.35");
	if (me == match && addr == match2) {
	  struct timeval tv;
	  click_gettimeofday(&tv);
	  click_chatter("* hello: neighbor %s (@ %s, %d sec %d usec), lifetime is %d\n", addr.s().cc(), me.s().cc(),
			tv.tv_sec, tv.tv_usec, route->lifetime_jiffies - gap);
	}
	// timeout?  entries with lifetimes of -1 don't time out...
	if (route->lifetime_jiffies > -1) {
	  if (route->hop_count < RT_INFINITY) {
	    if (gap > route->lifetime_jiffies) {
	      if (route == info->shortest_route) {
		info->shortest_route = NULL;
		info->last_hop_count = route->hop_count;
	      }
	      if (route == info->shortest_source_route) {
		info->shortest_source_route = NULL;
		info->last_source_hop_count = route->hop_count;
	      }
	      
	      struct timeval tv;
	      click_gettimeofday(&tv);
	      click_chatter("* rt: expired route %s (@ %s, %d sec %d usec)\n", addr.s().cc(), me.s().cc(),
			    tv.tv_sec, tv.tv_usec);
	      if (route->hop_count == 0) {
		// next hop; have to expire all routes using this node
		inc_seq = true;
		dst_list.push_back(info->dst);
	      }
	      route->hop_count = RT_INFINITY;
	      route->last_response_jiffies = jiff;
	      _route_breaks++;
	      route_break = true;
	    } else {
	      if (gap < shortest_gap)
		shortest_gap = gap;
	    }
	  } else if (gap > DELETE_PERIOD_JIFFIES) {
	    click_chatter("* rt: route %s deleted (@ %s)\n", addr.s().cc(), me.s().cc());
	    // delete route	  
	    if (!prev_route)
	      info->rlist_head = route->next;
	    else
	      prev_route->next = route->next;
	    delete route;
	    route = NULL;
	  }
	}
	prev_route = route;
	route = next_route;
      }
      // recalculate new shortest route?
      if (!info->shortest_route) {
	find_shortest_route(info);
	// only increment sequence if we used to have a shortest route
	if (!info->shortest_route && prev_shortest_route)
	  // destination lost, so increment sequence number
	  info->dst_seq++;
      }
      if (!info->shortest_source_route)
	find_shortest_source_route(info);
      // release lock
      info->lock->release();
    }
  }

  _table_lock->release();
  // new sequence number?
  if (inc_seq)
    _seq++;
  // link break counter
  if (route_break)
    _link_breaks++;
  // invalidate additional links as result of link break?
  if (dst_list.size() > 0)
    invalidate_neighbor_link(dst_list);

  // todo: remove
  _expire_timer.schedule_after_ms((1000 * shortest_gap) / CLICK_HZ);
}

// warning: acquire lock on route info before calling
void
RoutingTable::find_shortest_route(RouteInfo *info)
{
  info->shortest_route = NULL;
  info->shortest_source_route = NULL;
  RouteEntry *route = info->rlist_head;

  // recalculate new best route for this destination
  while(route) {
    if (route->hop_count < RT_INFINITY && 
	(!info->shortest_route || route->hop_count < info->shortest_route->hop_count)) {
      // found a shorter route
      info->shortest_route = route;
    }
    route = route->next;
  }
}

// warning: acquire lock on route info before calling
void
RoutingTable::find_shortest_source_route(RouteInfo *info)
{
  info->shortest_source_route = NULL;
  RouteEntry *route = info->rlist_head;

  // recalculate new shortest route for this destination
  while(route) {
    if (route->source_route && route->hop_count < RT_INFINITY &&
	(!info->shortest_source_route || route->hop_count < info->shortest_source_route->hop_count)) {
      // found a shorter source route
      info->shortest_source_route = route;
    }
    route = route->next;
  }
}

int
RoutingTable::find_shortest_lifetime(int jiff, int &shortest_lifetime_jiffies)
{
  RouteInfo *info;
  unsigned dst = 0;
  
  shortest_lifetime_jiffies = -1;

  // lock routing table
  _table_lock->acquire();

  for (int i = 0; i < _routes; i++) {
    info = _as[i];
    if (info && info->polling) {
      int my_lifetime_jiffies = 
	info->poll_lifetime_jiffies - (jiff - info->last_poll_jiffies);
      if (shortest_lifetime_jiffies == -1 || 
	  my_lifetime_jiffies < shortest_lifetime_jiffies) {
	shortest_lifetime_jiffies = my_lifetime_jiffies;
	dst = info->dst;
      }
    }
  }

  _table_lock->release();

  return dst;
}

void
RoutingTable::get_nodes(Vector<unsigned> &dst_list, unsigned hop_count)
{
  // get all nodes within the hop_count radius from us
  _table_lock->acquire();

  for (int i = 0; i < _routes; i++) {
    RouteInfo *info = _as[i];
    if (info) {
      info->lock->acquire();
      RouteEntry *route = info->rlist_head;
      while (route) {
	if (route->hop_count <= hop_count)
	  dst_list.push_back(info->dst);
	route = route->next;
      }
      info->lock->release();
    }
  }

  _table_lock->release();
}

// warning: acquire lock on route info before calling
void
RoutingTable::add_p_entry(RouteInfo* info, Packet *p)
{
  PacketEntry *p_entry = new PacketEntry;

  // add to end of packetlist
  if (!info->plist_head) {
    info->plist_head = p_entry;
    info->p_entries = 1;
  } else {
    info->plist_tail->next = p_entry;
    info->p_entries++;
  }
  info->plist_tail = p_entry;
  p_entry->p = p;
  p_entry->next = NULL;
}

// warning: acquire lock before calling
PacketEntry*
RoutingTable::remove_p_entries(RouteInfo* info)
{
  PacketEntry *p_entry = NULL;

  if (info) {
    p_entry = info->plist_head;
    info->plist_head = NULL;
    info->plist_tail = NULL;
    info->p_entries = 0;
  }

  return p_entry;
}

bool
RoutingTable::interface_ip(unsigned src, int index)
{
  unsigned ip;

  if (index > -1 && index < _ip_iflist.size()) {
    if ((ip = _ip_iflist[index]) == src)
      return true;
  } else
    click_chatter("* rt: invalid interface\n");
    
  return false;
}

IPAddress
RoutingTable::interface_ip(int index)
{
  // return address stored for interface 'index'
  if (index > -1 && index < _ip_iflist.size())
    return _ip_iflist[index];
  else
    click_chatter("* rt: invalid interface\n");

  return 0;
}

EtherAddress
RoutingTable::interface_ether(int index)
{
  // return address stored for interface 'index'

  if (index < _ether_iflist.size())
    return _ether_iflist[index];
  else
    click_chatter("*rt: invalid interface\n");

  return EtherAddress(0);
}

bool
RoutingTable::any_interface(unsigned src)
{
  // check to see if we have any interfaces registered with ip address 'src'
  for (int i = 0; i < _ip_iflist.size(); i++) {
    unsigned ip = _ip_iflist[i];
    if (ip == src)
      return true;
  }

  return false;
}

bool
RoutingTable::any_interface(struct in_addr ip_src)
{
  // check to see if we have any interfaces registered with ip address 'src'
  for (int i = 0; i < _ip_iflist.size(); i++) {
    if (_ip_iflist[i] == ip_src)
      return true;
  }

  return false;
}

void
RoutingTable::set_sequence(unsigned int adhoc_seq)
{
  _seq = adhoc_seq;
}

int
RoutingTable::last_hello_jiffies()
{
  return _last_hello_jiffies;
}

void
RoutingTable::update_hello()
{
  _last_hello_jiffies = click_jiffies();
}

void
RoutingTable::add_info(unsigned dst)
{
  RouteInfo *info = create_and_acquire_info(dst);

  if (info) {
    release(info);
  }
}

void
RoutingTable::add_route(unsigned dst, unsigned char *ethernet, unsigned interface, int lifetime_jiffies)
{
  // add next hop (final destination, no intermediate hops) route

  RouteInfo *info = create_and_acquire_info(dst);
  RouteEntry *route = create_neighbor_route(info, lifetime_jiffies);

  if (info && route) {
    info->interface = interface;
    memcpy(info->ethernet, ethernet, 6);

    IPAddress addr(dst);
    EtherAddress ether(ethernet);
    click_chatter("* rt: added route for %s, ethernet is %s, interface is %d\n",
		  addr.s().cc(), ether.s().cc(), info->interface);
  } else
    click_chatter("* rt: error adding route\n");

  release(info);
}

void
RoutingTable::add_route(unsigned dst, unsigned next_hop, unsigned hop_count, 
			unsigned dst_seq, int lifetime_jiffies)
{
  // add next hop route
  Vector<unsigned> hops;

  RouteInfo *info = create_and_acquire_info(dst);
  RouteEntry *route = NULL;

  if (info) {
    if (hop_count == 0) {
      // next hop is destination, no intermediate hops
      route = create_neighbor_route(info, lifetime_jiffies);
      // if we don't know the interface and link layer address
      // of neighbor, we can't use it!
      if (info->interface == RT_UNKNOWN_IF)
	route->hop_count = RT_INFINITY;
      else 
	info->polling = false;
    } else {
      // next hop is not final destination
      click_chatter("* aodv: creating next hop route...\n");
      route = create_next_hop_route(info, next_hop, hop_count, lifetime_jiffies);
      /*
      hops.resize(2);
      hops[0] = next_hop;
      hops[1] = dst;
      route = create_route(info, hops, hop_count, (CLICK_HZ * lifetime_ms) / 1000);
      */
      // add route info data
      info->polling = false;
    } 

    if (dst_seq > 0)
      info->dst_seq = dst_seq;
      
    IPAddress dst_addr(dst);
    IPAddress next(next_hop);
    click_chatter("* rt: added route for %s, next hop is %s, "
		  "hop count is %d, lifetime is %d\n",
		  dst_addr.s().cc(), next.s().cc(), hop_count, lifetime_jiffies);
  }
}

void
RoutingTable::add_route(const Vector<unsigned> &hops)
{
  int hop_count = hops.size() - 1;

  if (hop_count >= 0) {
    // last entry is destination
    RouteInfo *info = create_and_acquire_info(hops[hop_count]);
    RouteEntry *route = NULL;

    if (info) {
      if (hop_count == 0) {
	// next hop is destination, no intermediate hops
	route = create_neighbor_route(info, _source_lifetime_jiffies);
	if (info->interface == RT_UNKNOWN_IF)
	  route->hop_count = RT_INFINITY;
	else
	  info->polling = false;
      } else {
	// remove destination hop
	route = create_route(info, hops, hop_count, _source_lifetime_jiffies);
	// add route info data
	info->polling = false;
      }
      IPAddress dst_addr(info->dst);
      IPAddress next(hops[0]);
      click_chatter("* rt: added route for %s, next hop is %s, hop count is %d\n",
		    dst_addr.s().cc(), next.s().cc(), route->hop_count);
      // release route info lock
      release(info);
    } else
      click_chatter("* rt: could not add route info\n");
  } else
    click_chatter("* rt: could not add route - hop count is zero\n");
}

void
RoutingTable::refresh_route(unsigned dst, bool source)
{
  RouteInfo *info = acquire(dst);
  RouteEntry *route = NULL;

  if (info) {
    // todo: should really refresh correct route
    // (could possibly be different from shortest route)
    if (source)
      route = info->shortest_source_route;
    else
      route = info->shortest_route;
    
    if (route && route->hop_count < RT_INFINITY) {
      /*
      if (source)
	route->lifetime_jiffies = _source_lifetime_jiffies;
      else
	route->lifetime_jiffies = _lifetime_jiffies;
      */
      // update last response
      //if (route->hop_count > 0 && route->hops[0] != dst) {
	route->last_response_jiffies = click_jiffies();
	// don't have to wait for an ack anymore
	info->areq = false;
	//}
    }
    release(info);
  }
}

void
RoutingTable::add_precursor_info(unsigned dst, unsigned last_hop)
{
  RouteInfo *info = create_and_acquire_info(dst);

  if (info && last_hop) {
    // todo: make sure we didn't already add this destination
    IPAddress addr(last_hop);
    IPAddress dst(dst);
    //info->precursor_map.insert(addr.s(), 1);
  }
  else
    click_chatter("* rt: could not add precursor info\n");

  //release(info);
}

void
RoutingTable::invalidate_neighbor_link(Vector<unsigned> &dst_list)
{
  unsigned dst;

  for (int i = 0; i < _routes; i++) {
    RouteInfo *info = _as[i];
    if (info) {
      RouteEntry *prev_shortest_route = info->shortest_route;
      for (int j = 0; j < dst_list.size(); j++) {
	RouteEntry *route = info->rlist_head;
	while (route) {
	  if (route->hop_count < RT_INFINITY) {
	    if (route->hop_count > 0)
	      dst = route->hops[0];
	    else
	      dst = info->dst;
	    if (dst == dst_list[j]) {
	      IPAddress addr(info->dst);
	      click_chatter("* rt: invalidated route to %s\n", addr.s().cc());
	      dst_list.push_back(info->dst);
	      // no longer best route
	      if (info->shortest_route == route) {
		info->shortest_route = NULL;
		info->last_hop_count = route->hop_count;
	      }
	      if (info->shortest_source_route == route) {
		info->shortest_source_route = NULL;
		info->last_source_hop_count = route->hop_count;
	      }
	      route->hop_count = RT_INFINITY;
	      route->last_response_jiffies = click_jiffies();
	    }
	  }
	  route = route->next;
	}
      }
      // recalculate new shortest route?
      if (!info->shortest_route) {
	find_shortest_route(info);
	if (!info->shortest_route && prev_shortest_route)
	  // destination lost, so increment sequence number
	  info->dst_seq++;
      }
      if (!info->shortest_source_route)
	find_shortest_source_route(info);      
    }
  }

  /*
  _table_lock->acquire();

  for (int i = 0; i < dst_list.size(); i++) {
    for (int j = 0; j < _routes; j++) {
      RouteInfo *info = _as[j];
      if (info) {
	RouteEntry *route = info->rlist_head;
	while (route) {
	  if (route->hop_count < RT_INFINITY) {
	    if (route->hop_count > 0)
	      dst = route->hops[0];
	    else
	      dst = info->dst;
	    if (dst == dst_list[i]) {
	      route->hop_count = RT_INFINITY;
	      route->last_response_jiffies = click_jiffies();
	      IPAddress addr(info->dst);
	      click_chatter("* rt: invalidated route to %s\n", addr.s().cc());
	      dst_list.push_back(info->dst);
	    }
	  }
	  route = route->next;
	}
      }
    }
  }

  _table_lock->release();
  */
}

void
RoutingTable::invalidate_link(unsigned src, unsigned dst)
{
  unsigned route_dst, hop_count;
  bool route_break = false;

  _table_lock->acquire();

  for (int i = 0; i < _routes; i++) {
    RouteInfo *info = _as[i];
    if (info) {
      RouteEntry *route = info->rlist_head;
      while (route) {
	unsigned index = 0;
	while (index < (unsigned)route->hops.size()) {
	  if (route->hops[index] == src) {
	    if (index + 1 < route->hop_count)
	      route_dst = route->hops[index + 1];
	    else
	      route_dst = info->dst;
	    if (route_dst == dst) {
	      // link is broken, invalidate route
	      hop_count = route->hop_count;
	      route->hop_count = RT_INFINITY;
	      _route_breaks++;
	      route_break = true;
	      IPAddress addr_src(src);
	      IPAddress addr_dst(dst);
	      click_chatter("* rt: invalidated link from %s to %s\n",
			    (char*)addr_src.s().data(), (char*)addr_dst.s().data());
	      index = route->hop_count;
	      if (info->shortest_route == route) {
		info->shortest_route = NULL;
		info->last_hop_count = hop_count;
	      }
	      if (info->shortest_source_route == route) {
		info->shortest_source_route = NULL;
		info->last_source_hop_count = route->hop_count;
	      }
	    }
	  }
	  index++;
	}
	route = route->next;
      }
      if (!info->shortest_route) {
	// destination lost, so increment sequence number
	info->dst_seq++;
	find_shortest_route(info);
      }
      if (!info->shortest_source_route)
	find_shortest_source_route(info);
    }
  }
  _table_lock->release();

  if (route_break)
    _link_breaks++;
}

// warning: info must be locked prior to calling this function
void
RoutingTable::invalidate_route(RouteInfo *info, RouteEntry *route, unsigned dst_seq)
{
  if (info && route) {
    unsigned hop_count = route->hop_count;
    
    route->hop_count = RT_INFINITY;
    info->dst_seq = dst_seq;
    _route_breaks++;
    
    if (route == info->shortest_route) {
      info->last_hop_count = hop_count;
      info->shortest_route = NULL;
      find_shortest_route(info);
    }
    if (route == info->shortest_source_route) {
      info->last_source_hop_count = hop_count;
      info->shortest_source_route = NULL;
      find_shortest_source_route(info);
    }
    IPAddress addr(info->dst);
    click_chatter("* rt: invalidated route to %s\n", addr.s().cc());
  }
}

// warning: info must be locked prior to calling this function
RouteEntry*
RoutingTable::create_route(RouteInfo *info, const Vector<unsigned> &hops, unsigned hop_count, int lifetime_jiffies)
{
  RouteEntry *route = NULL;
  unsigned vector_hops = hops.size() - 1;

  if (info) {
    RouteEntry *next_route = info->rlist_head;
    
    // check to see if we have an identical route already
    while (!route && next_route) {
      if (next_route->hop_count == vector_hops) {
	unsigned index = 0;
	bool same = true;
	while (same && index < vector_hops) {
	  if (next_route->hops[index] != hops[index])
	    same = false;
	  index++;
	}
	if (same) {
	  route = next_route;
	  IPAddress addr(info->dst);
	  click_chatter("* rt: updating route entry %s\n", addr.s().cc());
	}
      }
      next_route = next_route->next;
    }
    
    if (!route) {
      route = new RouteEntry;
      assert(route);
      route->next = info->rlist_head;
      info->rlist_head = route;
      route->hops.resize(vector_hops);
      // add route hops
      for (unsigned i = 0; i < vector_hops; i++) {
	route->hops[i] = hops[i];
      }
      route->source_route = false;
      route->hop_count = hop_count;
      // new shortest route?
      if (route->hop_count < RT_INFINITY) {
	if (!info->shortest_route || route->hop_count <= info->shortest_route->hop_count) {
	  if (info->shortest_route)
	    info->last_hop_count = info->shortest_route->hop_count;
	  info->shortest_route = route;
	  // source route?
	  if (hop_count == vector_hops) {
	    route->source_route = true;
	    // new shortest source route?
	    if (!info->shortest_source_route || route->hop_count <= info->shortest_source_route->hop_count) {
	      if (info->shortest_source_route)
		info->last_source_hop_count = info->shortest_source_route->hop_count;
	      info->shortest_source_route = route;
	    }
	  }
	}
      }
      IPAddress addr(info->dst);
      click_chatter("* rt: added new route to %s\n", addr.s().cc());
    }
    // refresh entry
    route->lifetime_jiffies = lifetime_jiffies;
    route->last_response_jiffies = click_jiffies();
    info->areq = false;
  }

  return route;
}

// warning: info must be locked prior to calling this function
RouteEntry*
RoutingTable::create_neighbor_route(RouteInfo *info, int lifetime_jiffies)
{
  RouteEntry *route = NULL;

  if (info) {
    RouteEntry *next_route = info->rlist_head;
    
    // check to see if we have a neighbor, no hop route already
    while (!route && next_route) {
      if (next_route->hop_count == 0) {
	  route = next_route;
	  IPAddress addr(info->dst);
	  click_chatter("* rt: updating route entry %s\n", addr.s().cc());
      } else
	next_route = next_route->next;
    }
    
    if (!route) {
      route = new RouteEntry;
      assert(route);
      route->next = info->rlist_head;
      info->rlist_head = route;
      route->hop_count = 0;
      route->source_route = true; // complete route path

      IPAddress addr(info->dst);
      click_chatter("* rt: added new route to %s\n", addr.s().cc());
    }
    // definitely shortest route
    info->shortest_route = route;
    info->shortest_source_route = route;
    // refresh entry
    route->lifetime_jiffies = lifetime_jiffies;
    route->last_response_jiffies = click_jiffies();
    info->areq = false;
  }

  return route;
}

// warning: info must be locked prior to calling this function
RouteEntry*
RoutingTable::create_next_hop_route(RouteInfo *info, unsigned next_hop, unsigned hop_count, int lifetime_jiffies)
{
  RouteEntry *route = NULL;

  if (info) {
    RouteEntry *next_route = info->rlist_head;
    
    // check to see if we have a next hop route (only keeping one at the moment)
    while (!route && next_route) {
      if (next_route->hops.size() == 1) {
	  route = next_route;
	  IPAddress addr(info->dst);
	  IPAddress next(next_hop);
	  click_chatter("* rt: updating route entry %s, next hop is %s\n", addr.s().cc(), next.s().cc());
      } else
	next_route = next_route->next;
    }
    
    if (!route) {
      route = new RouteEntry;
      assert(route);
      route->next = info->rlist_head;
      info->rlist_head = route;
      route->hops.resize(1);

      IPAddress addr(info->dst);
      click_chatter("* rt: added new route to %s\n", addr.s().cc());
    }

    route->hops[0] = next_hop;
    route->hop_count = hop_count;
    route->source_route = false;

    // definitely shortest route
    //if (!info->shortest_route || info->shortest_route->hop_count > hop_count) {
      if (info->shortest_route)
	info->last_hop_count = info->shortest_route->hop_count;
      info->shortest_route = route;
      // }

    // refresh entry
    route->lifetime_jiffies = lifetime_jiffies;
    route->last_response_jiffies = click_jiffies();
    info->areq = false;
  }

  return route;
}

RouteInfo*
RoutingTable::create_and_acquire_info(unsigned dst)
{
  RouteInfo *info = NULL;
  IPAddress addr(dst);
  int index = _map[addr.s()];

  if (index >= 0)
    info = _as[index];

  if (!info) {
    // map this info entry
    info = new RouteInfo;
    assert(info);
    _routes++;
    info->dst_seq = 0;
    info->requests = 0;
    info->last_poll_jiffies = click_jiffies();
    info->dst = dst;
    info->ttl = 1;
    info->areq = false;
    info->polling = false;
    info->salvage = false;
    info->interface = RT_UNKNOWN_IF;
    info->lock = new Spinlock();
    info->last_hop_count = 0;
    info->last_source_hop_count = 0;
    info->shortest_route = NULL;
    info->shortest_source_route = NULL;
    info->rlist_head = NULL;
    info->plist_head = NULL;
    info->plist_tail = NULL;
    info->p_entries = 0;
    _map.insert(addr, _as.size());
    _as.push_back(info);
    click_chatter("* rt: %s added to index %d\n", addr.s().cc(), _as.size());
  }

  info->lock->acquire();

  return info;
}

RouteInfo*
RoutingTable::acquire(unsigned query_addr)
{
  IPAddress addr(query_addr);
  int index = _map[addr.s()];
  RouteInfo *info = NULL;
  
  if ((index >= 0) && (index < _as.size())) {
    // get table lock
    _table_lock->acquire();
    info = _as[index];
    if (info) {
      // get route entry lock
      info->lock->acquire();
    } else
      click_chatter("* rt: error: found route index, but no route pointer!\n");
    _table_lock->release();
  }

  if (!info)
    click_chatter("* rt: failed to acquire route %s (index %d)\n",
		  addr.s().cc(), index);

  return info;
}

void
RoutingTable::release(RouteInfo *info)
{
  if (info)
    info->lock->release();
}

void
RoutingTable::reset_breaks()
{
  _link_breaks = 0;
  _route_breaks = 0;
}

void
RoutingTable::packet_count()
{
  _table_lock->acquire();

  for (int i = 0; i < _routes; i++) {
    RouteInfo *info = _as[i];
    if (info) {
      IPAddress addr(info->dst);
      click_chatter("%d packets still queued for destination %s (@ %s)\n",
		    info->p_entries, addr.s().cc(), _ip_iflist[0].s().cc());
    }
  }
}

EXPORT_ELEMENT(RoutingTable)
ELEMENT_MT_SAFE(RoutingTable)
