#ifndef ROUTINGTABLE_HH
#define ROUTINGTABLE_HH

/*
 * =c
 * RoutingTable
 * =s routing table
 * static information element that stores the global routing table
 * =d
 * Stores all global routing table data so that it is accesible from all
 * elements in configuration.
 *
 * =e Instantiate at beginning of click configuration:
 *    // My interface
 *    AddressInfo(me0 192.168.1.115 00:03:47:70:89:14);
 *    // My routing table
 *    RoutingTable(me0);
 *
 * =a AODVAddRoute, DSRCache
 */

#include <click/element.hh>
#include <click/hashmap.hh>
#include <click/timer.hh>
#include <click/ipaddress.hh>
#include <click/etheraddress.hh>
#include <click/click_adhoc.hh>

#ifdef CLICK_SIM
namespace CLICKSIM {
#endif 

typedef void (*ExpireHook)(Timer *, void *);

class Spinlock;

class RoutingTable : public Element {

public:

  RoutingTable();
  ~RoutingTable();

  // element handling
  const char *class_name() const	{ return "RoutingTable"; }
  int configure_phase() const		{ return CONFIGURE_PHASE_FIRST; }
  RoutingTable *clone() const;
  int initialize(ErrorHandler *);
  void uninitialize();
  int configure(const Vector<String> &conf, ErrorHandler *errh);

  // routing table implementation
  int routes() { return _routes; }
  unsigned int sequence() { return _seq; }
  unsigned int route_breaks() { return _route_breaks; }
  unsigned int link_breaks() { return _link_breaks; }
  void reset_breaks();
  int source_route_lifetime() { return _source_lifetime_jiffies; }
  int route_lifetime() { return _lifetime_jiffies; }
  IPAddress interface_ip(int index);
  bool interface_ip(unsigned src, int index);
  EtherAddress interface_ether(int index);
  bool any_interface(unsigned src);
  bool any_interface(struct in_addr ip_src);
  void set_sequence(unsigned int seq);
  void update_hello();
  int last_hello_jiffies();
  void find_shortest_route(RouteInfo *info);
  void find_shortest_source_route(RouteInfo *info);
  int find_shortest_lifetime(int jiff, int &shortest_lifetime_jiffies);
  void clear_map();
  RouteInfo* acquire(unsigned query_addr);
  void release(RouteInfo *info);
  void add_p_entry(RouteInfo* info, Packet *p);
  PacketEntry* remove_p_entries(RouteInfo* info);
  void add_info(unsigned dst);
  void add_route(unsigned dst, unsigned char *ethernet, unsigned interface, int lifetime_jiffies);
  void add_route(unsigned dst, unsigned next_hop, unsigned hop_count, 
		 unsigned dst_seq, int lifetime_jiffies);
  void add_route(const Vector<unsigned> &hops);
  void refresh_route(unsigned dst, bool source);
  void add_precursor_info(unsigned dst, unsigned last_hop);
  void invalidate_link(unsigned src, unsigned dst);
  void invalidate_neighbor_link(Vector<unsigned> &dst_list);
  void invalidate_route(RouteInfo *info, RouteEntry *route, unsigned dst_seq);
  void get_nodes(Vector<unsigned> &dst_list, unsigned hop_count);

  void register_expire_callback(ExpireHook hook, Vector<unsigned> *dst_list, 
				void *thunk);
  static void expire_hook(Timer *, void *);
  void do_expire_callbacks(Timer *timer, Vector<unsigned> &dst_list);

private:
  void packet_count();
  
  // route expiration callbacks
  void expire_routes(Vector<unsigned> &dst_list);

  // route creation
  RouteInfo* create_and_acquire_info(unsigned dst);
  RouteEntry* create_route(RouteInfo *info, const Vector<unsigned> &hops, 
			   unsigned hop_count, int lifetime_jiffies);
  RouteEntry* create_neighbor_route(RouteInfo *info, int lifetime_jiffies);
  RouteEntry* create_next_hop_route(RouteInfo *info, unsigned next_hop,
				    unsigned hop_count, int lifetime_jiffies);

  struct ExpireEntry {
    ExpireHook hook;
    Vector<unsigned> *dst_list;
    void *thunk;
  };
  
  Timer _expire_timer;
  Vector<ExpireEntry*> _expire_callbacks;
  HashMap<String, int> _map;
  Vector<RouteInfo*> _as;
  Vector<IPAddress> _ip_iflist;
  Vector<EtherAddress> _ether_iflist;
  int _routes; // number of routes
  unsigned int _seq; // current source sequence number
  int _last_hello_jiffies;
  int _route_breaks;
  int _link_breaks;
  int _lifetime_jiffies;
  int _source_lifetime_jiffies;
  short int _id;
  // anyone that make changes to the _map lookup table need to
  // acquire a lock first
  Spinlock *_table_lock;

  static const int DELETE_PERIOD_JIFFIES = (CLICK_HZ * DELETE_PERIOD) / 1000;
};

#ifdef CLICK_SIM
}
#endif 

#endif
